#include <sys/os32.h>
#include <kernel/dev.h>
#include <sys/types.h>
#include <errno.h>
#include <kernel/memory.h>
#include <kernel/proc.h>
#include <kernel/net/arp.h>
#include <kernel/net/route.h>
#include <kernel/net.h>
#include <string.h>
#include <signal.h>
#include <ctype.h>
//----------------------------------------------------------
ssize_t
dev_kmem (pid_t pid, dev_t device, int rw,
          off_t offset, void *buffer, size_t size, int *eof)
{
  inode_t *inode;
  sb_t *sb;
  file_t *file;
  void *start;
  char *m;
  // 
  // Only read is allowed.
  // 
  if (rw != DEV_READ)
    {
      errset (EIO);     // I/O error.
      return ((ssize_t) - 1);
    }
  // 
  // Only positive offset is allowed.
  // 
  if (offset < 0)
    {
      errset (EIO);     // I/O error.
      return ((ssize_t) - 1);
    }
  // 
  // Read is selected (and is the only access
  // allowed).
  // 
  switch (device)
    {
    case DEV_KMEM_PS:
      // 
      // Verify if the selected slot can be read.
      // 
      if (offset >= PROCESS_MAX)
        {
          errset (EIO); // I/O error.
          return ((ssize_t) - 1);
        }
      // 
      // Correct the size to be read.
      // 
      if (sizeof (proc_t) < size)
        {
          size = sizeof (proc_t);
        }
      // 
      // Get the pointer to the selected slot.
      // 
      start = proc_reference ((pid_t) offset);
      break;
    case DEV_KMEM_MMP:
      // 
      // Correct the size to be read.
      // 
      if (offset >= (MEM_MAX_BLOCKS / 8))
        {
          *eof = 1;
          errset (EIO); // I/O error.
          return ((ssize_t) - 1);
        }
      // 
      // Reduce size if necessary.
      // 
      if ((offset + size) > (MEM_MAX_BLOCKS / 8))
        {
          size = ((MEM_MAX_BLOCKS / 8) - offset);
        }
      // 
      // Get the pointer to the map: offset is not
      // taken
      // into consideration.
      // 
      m = (char *) mb_reference ();
      // 
      start = &m[offset];
      // 
      break;
    case DEV_KMEM_SB:
      // 
      // Verify if the selected slot can be read.
      // 
      if (offset >= SB_MAX_SLOTS)
        {
          errset (EIO); // I/O error.
          return ((ssize_t) - 1);
        }
      // 
      // Get a reference to the super block table.
      // 
      sb = sb_reference (0);
      // 
      // Correct the size to be read.
      // 
      if (sizeof (sb_t) < size)
        {
          size = sizeof (sb_t);
        }
      // 
      // Get the pointer to the selected super block
      // slot.
      // 
      start = &sb[offset];
      break;
    case DEV_KMEM_INODE:
      // 
      // Verify if the selected slot can be read.
      // 
      if (offset >= INODE_MAX_SLOTS)
        {
          errset (EIO); // I/O error.
          return ((ssize_t) - 1);
        }
      // 
      // Get a reference to the inode table.
      // 
      inode = inode_reference (0, 0);
      // 
      // Correct the size to be read.
      // 
      if (sizeof (inode_t) < size)
        {
          size = sizeof (inode_t);
        }
      // 
      // Get the pointer to the selected inode slot.
      // 
      start = &inode[offset];
      break;
    case DEV_KMEM_FILE:
      // 
      // Verify if the selected slot can be read.
      // 
      if (offset >= FILE_MAX_SLOTS)
        {
          errset (EIO); // I/O error.
          return ((ssize_t) - 1);
        }
      // 
      // Get a reference to the file table.
      // 
      file = file_reference (0);
      // 
      // Correct the size to be read.
      // 
      if (sizeof (file_t) < size)
        {
          size = sizeof (file_t);
        }
      // 
      // Get the pointer to the selected inode slot.
      // 
      start = &file[offset];
      break;
    case DEV_KMEM_ARP:
      // 
      // Verify if the selected slot can be read.
      // 
      if (offset >= ARP_MAX_ITEMS)
        {
          errset (EIO); // I/O error.
          return ((ssize_t) - 1);
        }
      // 
      // Correct the size to be read.
      // 
      if (sizeof (arp_t) < size)
        {
          size = sizeof (arp_t);
        }
      // 
      // Get the pointer to the selected ARP item.
      // 
      start = &arp_table[offset];
      break;
    case DEV_KMEM_NET:
      // 
      // Verify if the selected slot can be read.
      // 
      if (offset >= NET_MAX_DEVICES)
        {
          errset (EIO); // I/O error.
          return ((ssize_t) - 1);
        }
      // 
      // Correct the size to be read.
      // 
      if (sizeof (net_t) < size)
        {
          size = sizeof (net_t);
        }
      // 
      // Get the pointer to the selected NET table
      // item.
      // 
      start = &net_table[offset];
      break;
    case DEV_KMEM_ROUTE:
      // 
      // Verify if the selected slot can be read.
      // 
      if (offset >= ROUTE_MAX_ROUTES)
        {
          errset (EIO); // I/O error.
          return ((ssize_t) - 1);
        }
      // 
      // Correct the size to be read.
      // 
      if (sizeof (route_t) < size)
        {
          size = sizeof (route_t);
        }
      // 
      // Get the pointer to the selected NET table
      // item.
      // 
      start = &route_table[offset];
      break;
    default:
      errset (ENODEV);  // No such device.
      return ((ssize_t) - 1);
    }
  // 
  // At this point, data is ready to be copied to the
  // buffer.
  // 
  memcpy (buffer, start, size);
  // 
  // Return size read.
  // 
  return (size);
}
