#ifndef _KERNEL_DRIVER_ATA_H
#define _KERNEL_DRIVER_ATA_H    1
//----------------------------------------------------------
#include <stdint.h>
#include <sys/types.h>
#include <time.h>
//----------------------------------------------------------
//
// I/O ports, used to access ATA bus registers. These
// I/O ports are different for every ATA bus.
//
#define ATA0_DATA          0x1F0        // r/w
#define ATA0_FEATURE       0x1F1        // -/w
#define ATA0_ERROR         0x1F1        // r/-
#define ATA0_COUNT         0x1F2        // r/w
#define ATA0_LOW           0x1F3        // r/w
#define ATA0_MID           0x1F4        // r/w
#define ATA0_HIGH          0x1F5        // r/w
#define ATA0_DEVICE        0x1F6        // r/w
#define ATA0_COMMAND       0x1F7        // -/w
#define ATA0_STATUS        0x1F7        // r/- regular
                                        // status
#define ATA0_CONTROL       0x3F6        // -/w
#define ATA0_ALTERNATE     0x3F6        // w/-
                                        // alternate
                                        // status
//
#define ATA1_DATA          0x170        // r/w
#define ATA1_FEATURE       0x171        // -/w
#define ATA1_ERROR         0x171        // r/-
#define ATA1_COUNT         0x172        // r/w
#define ATA1_LOW           0x173        // r/w
#define ATA1_MID           0x174        // r/w
#define ATA1_HIGH          0x175        // r/w
#define ATA1_DEVICE        0x176        // r/w
#define ATA1_COMMAND       0x177        // -/w
#define ATA1_STATUS        0x177        // r/- regular
                                        // status
#define ATA1_CONTROL       0x376        // -/w
#define ATA1_ALTERNATE     0x376        // w/-
                                        // alternate
                                        // status
//
#define ATA2_DATA          0x1E8        // r/w
#define ATA2_FEATURE       0x1E9        // -/w
#define ATA2_ERROR         0x1E9        // r/-
#define ATA2_COUNT         0x1EA        // r/w
#define ATA2_LOW           0x1EB        // r/w
#define ATA2_MID           0x1EC        // r/w
#define ATA2_HIGH          0x1ED        // r/w
#define ATA2_DEVICE        0x1EE        // r/w
#define ATA2_COMMAND       0x1EF        // -/w
#define ATA2_STATUS        0x1EF        // r/- regular
                                        // status
#define ATA2_CONTROL       0x3E6        // -/w
#define ATA2_ALTERNATE     0x3E6        // w/-
                                        // alternate
                                        // status
//
#define ATA3_DATA          0x168        // r/w
#define ATA3_FEATURE       0x169        // -/w
#define ATA3_ERROR         0x169        // r/-
#define ATA3_COUNT         0x16A        // r/w
#define ATA3_LOW           0x16B        // r/w
#define ATA3_MID           0x16C        // r/w
#define ATA3_HIGH          0x16D        // r/w
#define ATA3_DEVICE        0x16E        // r/w
#define ATA3_COMMAND       0x16F        // -/w
#define ATA3_STATUS        0x16F        // r/- regular
                                        // status
#define ATA3_CONTROL       0x366        // -/w
#define ATA3_ALTERNATE     0x366        // w/-
                                        // alternate
                                        // status
//
// Status register flags (regular or alternate).
//
#define ATA_STATUS_BSY            0x80  // Busy
#define ATA_STATUS_DRDY           0x40  // Ready
#define ATA_STATUS_DF             0x20  // Drive Fault
#define ATA_STATUS_DRQ            0x08  // Data
                                        // request
#define ATA_STATUS_ERR            0x01  // Error
//
// Values to put to the device register
//
#define ATA_DEVICE_CHS          0x00
#define ATA_DEVICE_LBA          0x40
#define ATA_DEVICE_MASTER       0x00
#define ATA_DEVICE_SLAVE        0x10
//
// Values to put to the command register
//
#define ATA_COMMAND_IDENTIFY_DEVICE      0xEC
#define ATA_COMMAND_READ_SECTORS         0x20
#define ATA_COMMAND_WRITE_SECTORS        0x30
#define ATA_COMMAND_FLUSH_CACHE          0xE7
//
// Values to put to the control register
// (device control register).
//
#define ATA_CONTROL_HOB        0x80
#define ATA_CONTROL_SRST       0x04     // Software
                                        // reset.
#define ATA_CONTROL_NIEN       0x01     // No
                                        // Interrupt
                                        // enabled.
//
//
//
#define ATA_MAX_DEVICES        8        // Fixed.
#define ATA_SECTOR_SIZE        512      // Fixed.
#define ATA_TIMEOUT  \
    ((clock_t) (CLOCKS_PER_SEC * 1))    // 1 s
#define ATA_TIMEOUT_FLUSH \
    ((clock_t) (CLOCKS_PER_SEC * 10))   // 10 s
//
//
//
typedef struct
{
  unsigned short r_data;
  unsigned short r_feature;
  unsigned short r_error;
  unsigned short r_count;
  unsigned short r_low;
  unsigned short r_mid;
  unsigned short r_high;
  unsigned short r_device;
  unsigned short r_command;
  unsigned short r_status;
  unsigned short r_control;
  unsigned short r_alternate;
  unsigned char bus;
  unsigned char target;
  unsigned char present;
  uint16_t id[ATA_SECTOR_SIZE / 2];
  unsigned int sectors;
} ata_t;
//
typedef struct
{
  char byte[ATA_SECTOR_SIZE];
} ata_sector_t;
//
extern ata_t ata_table[ATA_MAX_DEVICES];
//----------------------------------------------------------
void ata_init (void);
void ata_reset (int drive);
int ata_valid (int drive);
//----------------------------------------------------------
int ata_cmd_identify_device (int drive, void *buffer);
int ata_cmd_read_sectors (int drive,
                          unsigned int sector,
                          unsigned char count,
                          void *buffer);
int ata_cmd_write_sectors (int drive,
                           unsigned int sector,
                           unsigned char count,
                           void *buffer);
int ata_device (int drive, unsigned int sector);
//----------------------------------------------------------
int ata_rdy (int drive, clock_t timeout);
int ata_drq (int drive, clock_t timeout);
int ata_lba28 (int drive, unsigned int sector,
               unsigned char count);
//----------------------------------------------------------
#define ata_read_sector(drive, sector, buffer)  \
    (ata_cmd_read_sectors ((int) drive, \
     (unsigned int) sector, \
     (unsigned char) 1, (void *) buffer))
//
#define ata_write_sector(drive, sector, buffer)  \
    (ata_cmd_write_sectors ((int) drive, \
     (unsigned int) sector, \
     (unsigned char) 1, (void *) buffer))
//----------------------------------------------------------
#endif
