#ifndef _KERNEL_DRIVER_NIC_NE2K_H
#define _KERNEL_DRIVER_NIC_NE2K_H    1
//----------------------------------------------------------
#include <stdint.h>
#include <sys/types.h>
#include <unistd.h>
#include <time.h>
//----------------------------------------------------------
//
// Command register, available on all pages.
//
#define NE2K_CR         0x00    // rw
//
// Page 0 registers
//
#define NE2K_CLDA0      0x01    // r-
#define NE2K_PSTART     0x01    // -w
#define NE2K_CLDA1      0x02    // r-
#define NE2K_PSTOP      0x02    // -w
#define NE2K_BNRY       0x03    // rw
#define NE2K_TSR        0x04    // r-
#define NE2K_TPSR       0x04    // -w
#define NE2K_NCR        0x05    // r-
#define NE2K_TBCR0      0x05    // -w
#define NE2K_FIFO       0x06    // r-
#define NE2K_TBCR1      0x06    // -w
#define NE2K_ISR        0x07    // rw
#define NE2K_CRDA0      0x08    // r-
#define NE2K_RSAR0      0x08    // -w
#define NE2K_CRDA1      0x09    // r-
#define NE2K_RSAR1      0x09    // -w
#define NE2K_RBCR0      0x0A    // -w
#define NE2K_RBCR1      0x0B    // -w
#define NE2K_RSR        0x0C    // r-
#define NE2K_RCR        0x0C    // -w
#define NE2K_CNTR0      0x0D    // r-
#define NE2K_TCR        0x0D    // -w
#define NE2K_CNTR1      0x0E    // r-
#define NE2K_DCR        0x0E    // -w
#define NE2K_CNTR2      0x0F    // r-
#define NE2K_IMR        0x0F    // -w
//
// Page 1 registers: all read and write
//
#define NE2K_PAR0       0x01    // rw
#define NE2K_PAR1       0x02    // rw
#define NE2K_PAR2       0x03    // rw
#define NE2K_PAR3       0x04    // rw
#define NE2K_PAR4       0x05    // rw
#define NE2K_PAR5       0x06    // rw
#define NE2K_CURR       0x07    // rw
#define NE2K_MAR0       0x08    // rw
#define NE2K_MAR1       0x09    // rw
#define NE2K_MAR2       0x0A    // rw
#define NE2K_MAR3       0x0B    // rw
#define NE2K_MAR4       0x0C    // rw
#define NE2K_MAR5       0x0D    // rw
#define NE2K_MAR6       0x0E    // rw
#define NE2K_MAR7       0x0F    // rw
//
// Page 2 registers: read version of the same registers
// inside page 1:
//
//      NE2K_PSTART     0x01 // -w
//      NE2K_PSTOP      0x02 // -w
//      --              0x03
//      NE2K_TPSR       0x04 // -w
//      --              0x05
//      --              0x06
//      --              0x07
//      --              0x08
//      --              0x09
//      --              0x0A
//      --              0x0B
//      NE2K_RCR        0x0C // -w
//      NE2K_TCR        0x0D // -w
//      NE2K_DCR        0x0E // -w
//      NE2K_IMR        0x0F // -w
//
// Extra registers, outside pages.
//
#define NE2K_DATA       0x10    // Data port.
#define NE2K_RESET      0x1f    // Reset port.
//
// DMA buffer structure.
//
// 0x0000 .--------------------------.
//        |                          |
//        |                          |
//        |            ???           |
//        |                          |
//        |                          |
// 0x4000 |--------------------------|
//        | transmit buffer ring [1] |
// 0x4600 |--------------------------|
//        |                          |
//        |   receive buffer ring    |
//        |                          |
// 0xC000 |--------------------------|
//        |                          |
//        |           ???            |
//        |                          |
//        |                          |
//        `--------------------------' 0xFFFF
//
// [1] 0x600 is equal to 1536, that is the max space
//     that a packet can occupy inside the ring buffer.
//     So, there is the place for a
//     single transmission packet.
//
// Local DMA addresses:
//
#define NE2K_TX_START  0x40     // Means: 0x4000
#define NE2K_TX_STOP   0x46     // Means: 0x4600

#define NE2K_RX_START  0x46     // Means: 0x4600
#define NE2K_RX_STOP   0xC0     // Means: 0xC000
//
// Transmit buffer (remote DMA).
//
#define NE2K_TX_BUFFER NE2K_TX_START
//
// SA-PROM (Station Address PROM) size
//
#define NE2K_SAPROM_SIZE 32     // 32 bytes
//----------------------------------------------------------
int ne2k_check (uintptr_t io);
int ne2k_isr (uintptr_t io);
int ne2k_isr_expect (uintptr_t io, unsigned int isr_expect);
int ne2k_reset (uintptr_t io, void *address);
int ne2k_rx (uintptr_t io);
int ne2k_rx_reset (uintptr_t io);
int ne2k_tx (uintptr_t io, void *buffer, size_t size);
//
void ne2k_test2 (int eth);


//----------------------------------------------------------
#endif
