#include <kernel/driver/pci.h>
#include <kernel/net.h>
#include <kernel/driver/nic/ne2k.h>
#include <kernel/ibm_i386.h>
#include <errno.h>
#include <kernel/lib_k.h>
#include <kernel/lib_s.h>
//----------------------------------------------------------
#define DEBUG 0
//----------------------------------------------------------
int
ne2k_rx (uintptr_t io)
{
  int i;
  int bytes;
  int curr;
  int bnry;
  int next;
  int frame_status;
  int frame_size;
  int status;
  int n = net_index_eth (0, NULL, io);
  net_buffer_eth_t *buffer;
  // 
  // Verify to have found a valid Ethernet device.
  // 
  if (n < 0)
    {
      errset (ENODEV);
      return (-1);
    }
  // 
  // Command register (CR)
  // .-------------------------------.
  // |PS1|PS0|RD2|RD1|RD0|TXP|STA|STP|
  // |-------------------------------|
  // | 0 | 1 | 1 | 0 | 0 | 0 | 1 | 0 | 0x62
  // `-------------------------------'
  // \_____/ \_________/ |
  // | | START
  // | |
  // | Abort/complete remote DMA
  // |
  // Register page 1
  // 
  out_8 ((io + NE2K_CR), 0x62);
  // 
  // Get the current position.
  // 
  curr = in_8 (io + NE2K_CURR);
  // 
  // Command register (CR)
  // .-------------------------------.
  // |PS1|PS0|RD2|RD1|RD0|TXP|STA|STP|
  // |-------------------------------|
  // | 0 | 0 | 1 | 0 | 0 | 0 | 1 | 0 | 0x22
  // `-------------------------------'
  // \_____/ \_________/ |
  // | | START
  // | |
  // | Abort/complete remote DMA
  // |
  // Register page 0
  // 
  out_8 ((io + NE2K_CR), 0x22);
  // 
  // Get the boundary.
  // 
  bnry = in_8 (io + NE2K_BNRY);
  // 
  // ---------------------------------------------------
  // The function is run because at least a frame was
  // received:
  // if index `bnry' and index `curr' are the same,
  // all the receive
  // ring buffer is to be copied.
  // ---------------------------------------------------
  // 
  // Get all the frames ready from the internal
  // buffer.
  // 
  while (1)
    {
      // 
      // Find a place inside the frame table.
      // 
      buffer = net_buffer_eth (n);
      // 
      // Check to have a valid buffer pointer.
      // 
      if (buffer == NULL)
        {
          errset (errno);
          return (-1);
        }
      // 
      // First read 4 bytes starting from `bnry'.
      // 
      out_8 ((io + NE2K_RBCR0), 4);
      out_8 ((io + NE2K_RBCR1), 0);
      // 
      // Set the remote DMA address to bnry.
      // 
      out_8 ((io + NE2K_RSAR0), 0x00);  // Must be
      // zero.
      out_8 ((io + NE2K_RSAR1), bnry);
      // 
      // Command register (CR)
      // .-------------------------------.
      // |PS1|PS0|RD2|RD1|RD0|TXP|STA|STP|
      // |-------------------------------|
      // | 0 | 0 | 0 | 0 | 1 | 0 | 1 | 0 | 0x0A
      // `-------------------------------'
      // \_____/ \_________/ |
      // | | START
      // | |
      // | Read
      // |
      // Register page 0
      // 
      out_8 (io + NE2K_CR, 0x0A);
      // 
      // Frame status
      // 
      frame_status = in_8 (io + NE2K_DATA);
      // 
      // Next frame.
      // 
      next = in_8 (io + NE2K_DATA);
      // 
      // Frame size low.
      // 
      frame_size = in_8 (io + NE2K_DATA);
      // 
      // Frame size high
      // 
      frame_size += (in_8 (io + NE2K_DATA) * 256);
      // 
      // Interrupt status register (ISR)
      // .-------------------------------.
      // |RST|RDC|CNT|OVW|TXE|RXE|PTX|PRX|
      // |-------------------------------|
      // | 0 | 1 | 0 | 0 | 0 | 0 | 0 | 0 | 0x40
      // `-------------------------------'
      // |
      // Remote DMA complete
      // 
      // Verify to have finished with DMA transfer.
      // 
      status = ne2k_isr_expect (io, 0x40);
      if (status)
        {
          errset (errno);
          return (-1);
        }
      // 
      // Now read again all the frame plus header (the 
      // initial 4 bytes).
      // 
      buffer->clock = k_clock ();
      buffer->size = frame_size - 4;
      // 
      if (DEBUG)
        {
          k_printf
            ("0x%02x[BNRY=0x%02x "
             "CURR=0x%02x]0x%02x size=%i\n",
             NE2K_RX_START, bnry, curr, NE2K_RX_STOP,
             frame_size);
        }
      // 
      if (next == bnry)
        {
          k_printf
            ("[%s] next==bnry but should "
             "not happen!\n", __func__);
          errset (E_DRIVER_FAULT);
          return (-1);
        }
      // 
      if (next > bnry)
        {
          bytes = frame_size;
        }
      // 
      if (next < bnry)
        {
          // 
          // Read up to the bottom.
          // 
          bytes = ((NE2K_RX_STOP - bnry) * 256);
          bytes = min (bytes, frame_size);
        }
      // 
      // Read frame content: first part.
      // 
      out_8 ((io + NE2K_RBCR0), bytes & 0xFF);
      out_8 ((io + NE2K_RBCR1), bytes >> 8);
      // 
      out_8 ((io + NE2K_RSAR0), 0);     // MUST be
      // zero. :-(
      out_8 ((io + NE2K_RSAR1), bnry);
      // 
      // Command register (CR)
      // .-------------------------------.
      // |PS1|PS0|RD2|RD1|RD0|TXP|STA|STP|
      // |-------------------------------|
      // | 0 | 0 | 0 | 0 | 1 | 0 | 1 | 0 | 0x0A
      // `-------------------------------'
      // \_____/ \_________/ |
      // | | START
      // | |
      // | Read
      // |
      // Register page 0
      // 
      out_8 (io + NE2K_CR, 0x0A);
      // 
      // Jump the first four bytes (no way to start
      // after
      // the page start).
      // 
      in_8 (io + NE2K_DATA);
      in_8 (io + NE2K_DATA);
      in_8 (io + NE2K_DATA);
      in_8 (io + NE2K_DATA);
      bytes -= 4;
      // 
      // Get the frame data.
      // 
      i = 0;
      for (; bytes > 0; i++, bytes--)
        {
          buffer->frame.octet[i] = in_8 (io + NE2K_DATA);
        }
      // 
      // Interrupt status register (ISR)
      // .-------------------------------.
      // |RST|RDC|CNT|OVW|TXE|RXE|PTX|PRX|
      // |-------------------------------|
      // | 0 | 1 | 0 | 0 | 0 | 0 | 0 | 0 | 0x40
      // `-------------------------------'
      // |
      // Remote DMA complete
      // 
      // Verify to have finished with DMA transfer.
      // 
      status = ne2k_isr_expect (io, 0x40);
      if (status)
        {
          errset (errno);
          return (-1);
        }
      // 
      if (next < bnry)
        {
          // 
          // There might be a second part to read.
          // 
          bytes =
            frame_size - ((NE2K_RX_STOP - bnry) * 256);
        }
      // 
      if (bytes > 0)
        {
          // 
          out_8 ((io + NE2K_RBCR0), bytes & 0xFF);
          out_8 ((io + NE2K_RBCR1), bytes >> 8);
          // 
          out_8 ((io + NE2K_RSAR0), 0);
          out_8 ((io + NE2K_RSAR1), NE2K_RX_START);
          // 
          // Command register (CR)
          // .-------------------------------.
          // |PS1|PS0|RD2|RD1|RD0|TXP|STA|STP|
          // |-------------------------------|
          // | 0 | 0 | 0 | 0 | 1 | 0 | 1 | 0 | 0x0A
          // `-------------------------------'
          // \_____/ \_________/ |
          // | | START
          // | |
          // | Read
          // |
          // Register page 0
          // 
          out_8 (io + NE2K_CR, 0x0A);
          // 
          for (; bytes > 0; i++, bytes--)
            {
              buffer->frame.octet[i] =
                in_8 (io + NE2K_DATA);
            }
          // 
          // Interrupt status register (ISR)
          // .-------------------------------.
          // |RST|RDC|CNT|OVW|TXE|RXE|PTX|PRX|
          // |-------------------------------|
          // | 0 | 1 | 0 | 0 | 0 | 0 | 0 | 0 | 0x40
          // `-------------------------------'
          // |
          // Remote DMA complete
          // 
          // Verify to have finished with DMA
          // transfer.
          // 
          status = ne2k_isr_expect (io, 0x40);
          if (status)
            {
              errset (errno);
              return (-1);
            }
        }
      // 
      // Update BNRY.
      // 
      bnry = next;
      out_8 (io + NE2K_BNRY, bnry);
      // 
      // If the new bnry is equal to curr, the loop is 
      // finished.
      // 
      if (bnry == curr)
        {
          // 
          // finish.
          // 
          return (0);
        }
    }
}
