#include <kernel/driver/pci.h>
#include <kernel/driver/nic/ne2k.h>
#include <kernel/ibm_i386.h>
#include <errno.h>
#include <kernel/lib_k.h>
#include <kernel/lib_s.h>
//----------------------------------------------------------
int
ne2k_tx (uintptr_t io, void *buffer, size_t size)
{
  int i;
  int status;
  uint8_t *b = buffer;
  // 
  // Read the command register to see if the NIC is
  // transmitting.
  // The value 0x26 tells that the NIC is
  // transmitting.
  // 
  // Command register (CR)
  // .-------------------------------.
  // |PS1|PS0|RD2|RD1|RD0|TXP|STA|STP|
  // |-------------------------------|
  // | 0 | 0 | 1 | 0 | 0 | 1 | 1 | 0 | 0x26
  // `-------------------------------'
  // \_____/ \_________/ | |
  // | | | Start
  // | | Transmit frame
  // | Abort/complete
  // | remote DMA
  // Register
  // page 0
  // 
  status = in_8 (io + NE2K_CR);
  if (status == 0x26)
    {
      errset (EBUSY);
      return (-1);
    }
  // 
  // Set up the frame size: the size is split into
  // RBCR0 and RBCR1
  // registers.
  // 
  out_8 ((io + NE2K_RBCR0), (size & 0xFF));
  out_8 ((io + NE2K_RBCR1), (size >> 8));
  // 
  // Set the remote DMA address.
  // 
  out_8 ((io + NE2K_RSAR0), 0x00);      // Must be
  // zero.
  out_8 ((io + NE2K_RSAR1), NE2K_TX_BUFFER);
  // 
  // Command register (CR)
  // .-------------------------------.
  // |PS1|PS0|RD2|RD1|RD0|TXP|STA|STP|
  // |-------------------------------|
  // | 0 | 0 | 0 | 1 | 0 | 0 | 1 | 0 | 0x12
  // `-------------------------------'
  // \_____/ \_________/ |
  // | | Start
  // | |
  // | Write
  // Register
  // page 0
  // 
  out_8 ((io + NE2K_CR), 0x12);
  // 
  // Write to the data port all the frame.
  // 
  for (i = 0; i < size; i++)
    {
      out_8 ((io + NE2K_DATA), b[i]);
    }
  // 
  // Interrupt status register (ISR)
  // .-------------------------------.
  // |RST|RDC|CNT|OVW|TXE|RXE|PTX|PRX|
  // |-------------------------------|
  // | 0 | 1 | 0 | 0 | 0 | 0 | 0 | 0 | 0x40
  // `-------------------------------'
  // |
  // Remote DMA complete
  // 
  // Verify to have finished with DMA transfer.
  // 
  status = ne2k_isr_expect (io, 0x40);
  if (status)
    {
      errset (errno);
      return (-1);
    }
  // 
  // Set transmit page start, to the transmit buffer.
  // 
  out_8 (io + NE2K_TPSR, NE2K_TX_BUFFER);
  // 
  // Set transmit byte count (frame size).
  // 
  out_8 ((io + NE2K_TBCR0), (size & 0xFF));
  out_8 ((io + NE2K_TBCR1), (size >> 8));
  // 
  // Command register (CR)
  // .-------------------------------.
  // |PS1|PS0|RD2|RD1|RD0|TXP|STA|STP|
  // |-------------------------------|
  // | 0 | 0 | 1 | 0 | 0 | 1 | 1 | 0 | 0x26
  // `-------------------------------'
  // \_____/ \_________/ | |
  // | | | Start
  // | | Transmit frame
  // | Abort/complete remote DMA
  // Register
  // page 0
  // 
  // Send frame!
  // 
  out_8 ((io + NE2K_CR), 0x26);
  // 
  // Interrupt status register (ISR)
  // .-------------------------------.
  // |RST|RDC|CNT|OVW|TXE|RXE|PTX|PRX|
  // |-------------------------------|
  // | 0 | 0 | 0 | 0 | 1 | 0 | 1 | 0 | 0x0A
  // `-------------------------------'
  // | |
  // | Frame transmitted with no errors
  // Transmit error
  // 
  // Wait the end of transmission: might get a good
  // transmission
  // report, or an error transmission report.
  // 
  status = ne2k_isr_expect (io, 0x0A);
  if (status)
    {
      errset (errno);
      return (-1);
    }
  // 
  // Transmit status (TSR)
  // .-------------------------------.
  // |OWC|CDH| FU|CRS|ABT|COL| - |PTX|
  // |-------------------------------|
  // | 0 | 0 | 1 | 1 | 1 | 0 | 0 | 0 | 0x38
  // `-------------------------------'
  // | | |
  // | | Transmit aborted
  // | Carrier sense lost
  // FIFO underrun
  // 
  // Check if there was an error, during transmission.
  // 
  status = in_8 (io + NE2K_TSR);
  if (status & 0x38)
    {
      errset (EIO);
      return (-1);
    }
  // 
  // Done.
  // 
  return (0);
//
}
