#ifndef _KERNEL_FS_H
#define _KERNEL_FS_H 1
//----------------------------------------------------------
#include <stdint.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <limits.h>
#include <kernel/memory.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <kernel/net/ip.h>
#include <kernel/net/tcp.h>
#include <sys/os32.h>
//----------------------------------------------------------
#define SB_MAX_INODE_BLOCKS    8        // 8*8192
                                  // inodes max.
#define SB_MAX_ZONE_BLOCKS     8        // 8*8192
                                  // data-zones
                                  // max.
#define SB_BLOCK_SIZE       1024        // Fixed for
                                  // Minix file
                                  // system.
#define SB_MAX_ZONE_SIZE    4096        // log2 max is
                                  // 1.
//----------------------------------------------------------
//
// blocks * (1024 * 8 / 16)
//   = number of bits, divided 16.
//
#define SB_MAP_INODE_SIZE   (SB_MAX_INODE_BLOCKS*512)
#define SB_MAP_ZONE_SIZE    (SB_MAX_ZONE_BLOCKS*512)
//----------------------------------------------------------
//
// Number of zone pointers contained inside a zone,
// used as an indirect inode list
// (a pointer = 16 bits = 2 bytes).
//
#define INODE_MAX_INDIRECT_ZONES  (SB_MAX_ZONE_SIZE/2)
//----------------------------------------------------------
#define INODE_MAX_REFERENCES      0xFF
//----------------------------------------------------------
typedef uint16_t zno_t; // Zone number.
//----------------------------------------------------------
// The structured type `inode_t' must be pre-declared
// here, because the type sb_t, described before the
// inode structure, has a member pointing to a type
// `inode_t'. So, must be declared previously the type
// `inode_t' as made of a type `struct inode', then the
// structure `inode' can be described. But for a matter
// of coherence, all other structured data declared
// inside this file follow the same procedure.
//
typedef struct sb sb_t;
typedef struct inode inode_t;
typedef struct sock sock_t;
typedef struct file file_t;
typedef struct fd fd_t;
typedef struct directory directory_t;
//----------------------------------------------------------
#define SB_MAX_SLOTS  16        // Handle max 16 file
                                // systems.

struct sb
{       // File system super block:
  uint16_t inodes;      // inodes available;
  uint16_t zones;       // zones available (disk
  // size);
  uint16_t map_inode_blocks;    // inode bit map
  // blocks;
  uint16_t map_zone_blocks;     // data-zone bit map
  // blocks;
  uint16_t first_data_zone;     // first data-zone;
  uint16_t log2_size_zone;      // log_2
  // (size_zone/block_size);
  uint32_t max_file_size;       // max file size in
  // bytes;
  uint16_t magic_number;        // file system magic
  // number.
  // -------------------------------------------------
  // Extra management data, not saved inside the file
  // system
  // super block.
  // -------------------------------------------------
  dev_t device; // FS device [3]
  inode_t *inode_mounted_on;    // [4]
  blksize_t blksize;    // Calculated zone size.
  int options;  // [5]
  uint16_t map_inode[SB_MAP_INODE_SIZE];
  uint16_t map_zone[SB_MAP_ZONE_SIZE];
  char changed;
};

extern sb_t sb_table[SB_MAX_SLOTS];
//
// [3] the member `device' must be kept at the same
//     position, because it is used to calculate the
//     super block header size, saved on disk.
//
// [4] If this pointer is not NULL, the super block is
//     related to a device mounted on a directory. The
//     inode of such directory is recorded here. Please
//     note that it is type `void *', instead of type
//     `inode_t', because type `inode_t' is declared
//     after type `sb_t'.
//     Please note that the type `sb_t' is declared
//     before the type `inode_t', but this member
//     points to a type `inode_t'.
//     This is the reason because it was necessary to
//     declare first the type `inode_t' as made of
//     `struct inode', to be described later. For
//     coherence, all derived type made of structured
//     data, are first declared as structure, and then,
//     later, described.
//
// [5] Mount options can be only `MOUNT_DEFAULT' or
//     `MOUNT_RO', as defined inside file
//     `lib/sys/os32.h'.
//
//----------------------------------------------------------
#define INODE_MAX_SLOTS         (32 * OPEN_MAX)
#define INODE_PIPE_BUFFER_SIZE  18      // (7 dir. + 2
                                      // ind.) * 2.
//
struct inode
{       // Inode (32 byte total):
  uint16_t mode;        // file type and permissions;
  uint16_t uid; // user ID (16 bit);
  uint32_t size;        // file size in bytes;
  uint32_t time;        // file data modification
  // time;
  uint8_t gid;  // group ID (8 bit);
  uint8_t links;        // links to the inode;
  uint16_t direct[7];   // direct zones;
  uint16_t indirect1;   // indirect zones;
  uint16_t indirect2;   // double indirect zones.
  // -------------------------------------------------
  // Extra management data, not saved inside the disk
  // file system.
  // -------------------------------------------------
  sb_t *sb;     // Inode's super block. [7]
  ino_t ino;    // Inode number.
  sb_t *sb_attached;    // [8]
  blkcnt_t blkcnt;      // Rounded size/blksize.
  unsigned char references;     // Run time active
  // references.
  char changed:1,       // 1 == to be saved.
    pipe_dir:1; // 0 == read, 1 == write.
  unsigned char pipe_off_read;  // Pipe read offset.
  unsigned char pipe_off_write; // Pipe write offset 
  unsigned char pipe_ref_read;  // Pipe read
  // references.
  unsigned char pipe_ref_write; // Pipe write
  // references 
};

extern inode_t inode_table[INODE_MAX_SLOTS];
//
// [7] the member `sb' must be kept at the same
//     position, because it is used to calculate the
//     inode header size, saved on disk.
//
// [8] If the inode is a mount point for another
//     device, the other super block pointer is saved
//     inside `sb_attached'.
//
//----------------------------------------------------------
#define SOCK_MAX_SLOTS          64
#define SOCK_MAX_QUEUE          (SOCK_MAX_SLOTS/4)
//
struct sock
{
  int family;
  int type;
  int protocol;
  h_addr_t laddr;       // Local address, host byte
  // order.
  h_port_t lport;       // Local port, host byte
  // order.
  h_addr_t raddr;       // Remote address, host byte
  // order.
  h_port_t rport;       // Remote port, host byte
  // order.
  struct
  {
    clock_t clock[IP_MAX_PACKETS];      // [9]
  } read;
  uint8_t active:1,     // Is the socket used?
    unreach_net:1,      // 
    unreach_host:1,     // ICMP unreachable status.
    unreach_prot:1,     // 
    unreach_port:1;     // 
  struct
  {
    uint16_t conn:4,    // Connection status.
      can_write:1,      // Can write to send_data[].
      can_read:1,       // Can read from *recv_index.
      can_send:1,       // Can send data.
      can_recv:1,       // Can receive data.
      send_closed:1,    // Closed send direction.
      recv_closed:1;    // Closed receive direction.
    // 
    uint32_t lsq[16];   // Local sequence array.
    uint32_t lsq_ack;   // Expected acknowledge.
    uint32_t rsq[16];   // Remote sequence array.
    uint8_t lsqi:4,     // Local sequence array index.
      rsqi:4;   // Remote sequence array index.
    // 
    clock_t clock;      // When was last send.
    // 
    uint8_t send_data[TCP_MSS - sizeof (struct tcphdr)];
    size_t send_size;   // Size of `send_data[]'
    // content.
    int send_flags;
    uint8_t recv_data[TCP_MAX_DATA_SIZE];       // Data 
    // received.
    size_t recv_size;   // Size of `recv_data[]'
    // content.
    uint8_t *recv_index;        // Read index inside
    // `recv_data[]'.
    pid_t listen_pid;   // Process listening at local
    // port.
    int listen_max;     // Max connection requests.
    int listen_queue[SOCK_MAX_QUEUE];   // [10]
  } tcp;
};
//
extern sock_t sock_table[SOCK_MAX_SLOTS];
//
// [9] The array `read.clock[]' has the same size as
//     the array as `ip_tables[]', so that it can be
//     saved, inside the former, the clock time of a
//     packet read for the socket purposes.
//     This is necessary to know if the packet was
//     already managed inside the socket system, or
//     it is new.
//
// [10] When a process listen o a local port, member
//      `listen_pid' contains the pid number; member
//      `listen_max' contains the max allowed
//      connections that will be serviced; the array
//      `listen_queue[]' will contain the file
//      descriptors of established connections.
//      If `listen_queue[x]' is equal to -1, it means
//      that there is no file descriptor there.
//
//----------------------------------------------------------
#define FILE_MAX_SLOTS             (64 * OPEN_MAX)

struct file
{
  int references;
  off_t offset; // File position.
  int oflags;   // Open mode: r/w/r+w [11]
  inode_t *inode;
  sock_t *sock;
};

extern file_t file_table[FILE_MAX_SLOTS];
//
// [11] the member `oflags' can get only O_RDONLY,
//      O_WRONLY, O_RDWR, (from header `fcntl.h')
//      combined with OR binary operator.
//
//----------------------------------------------------------
struct fd
{
  int fl_flags; // File status flags and file
  // access modes. [12]
  int fd_flags; // File descriptor flags:
  // currently only FD_CLOEXEC.
  file_t *file; // Pointer to the file table.
};
//
// [12] the member `fl_flags' can get only O_RDONLY,
//      O_WRONLY, O_RDWR, O_CREAT, O_EXCL, O_NOCTTY,
//      O_TRUNC, O_APPEND and O_NONBLOCK
//      (from header `fcntl.h') combined with OR
//      binary operator.
//      Options like O_DSYNC, O_RSYNC and O_SYNC are
//      not taken into consideration by os32.
//
//  Please notice that each process has its own `fd'
//  table, embedded inside the process table.
//----------------------------------------------------------
struct directory
{       // Directory entry:
  uint16_t ino; // inode number;
  char name[NAME_MAX];  // file name.
};
//----------------------------------------------------------
void fs_init (void);
//----------------------------------------------------------
int sb_inode_status (sb_t * sb, ino_t ino);
sb_t *sb_mount (dev_t device, inode_t ** inode_mnt,
                int options);
void sb_print (void);
sb_t *sb_reference (dev_t device);
int sb_save (sb_t * sb);
int sb_zone_status (sb_t * sb, zno_t zone);
//----------------------------------------------------------
zno_t zone_alloc (sb_t * sb);
int zone_free (sb_t * sb, zno_t zone);
void zone_print (sb_t * sb, zno_t zone);
int zone_read (sb_t * sb, zno_t zone, void *buffer);
int zone_write (sb_t * sb, zno_t zone, void *buffer);
//----------------------------------------------------------
inode_t *inode_alloc (dev_t device, mode_t mode,
                      uid_t uid, gid_t gid);
int inode_check (inode_t * inode, mode_t type,
                 int perm, uid_t uid, gid_t gid);
int inode_dir_empty (inode_t * inode);
ssize_t inode_file_read (inode_t * inode, off_t offset,
                         void *buffer, size_t count,
                         int *eof);
ssize_t inode_file_write (inode_t * inode,
                          off_t offset,
                          const void *buffer, size_t count);
int inode_free (inode_t * inode);
blkcnt_t inode_fzones_read (inode_t * inode,
                            zno_t zone_start,
                            void *buffer, blkcnt_t blkcnt);
blkcnt_t inode_fzones_write (inode_t * inode,
                             zno_t zone_start,
                             void *buffer, blkcnt_t blkcnt);
inode_t *inode_get (dev_t device, ino_t ino);
inode_t *inode_pipe_make (void);
ssize_t inode_pipe_read (inode_t * inode, void *buffer,
                         size_t count, int *eof);
ssize_t inode_pipe_write (inode_t * inode,
                          const void *buffer, size_t count);
void inode_print (void);
int inode_put (inode_t * inode);
inode_t *inode_reference (dev_t device, ino_t ino);
int inode_save (inode_t * inode);
inode_t *inode_stdio_dev_make (dev_t device, mode_t mode);
int inode_truncate (inode_t * inode);
zno_t inode_zone (inode_t * inode, zno_t fzone, int write);
//----------------------------------------------------------
file_t *file_pipe_make (void);
file_t *file_reference (int fno);
file_t *file_stdio_dev_make (dev_t device, mode_t mode,
                             int oflags);
//----------------------------------------------------------
dev_t path_device (pid_t pid, const char *path);
int path_fix (char *path);
int path_full (const char *path,
               const char *path_cwd, char *full_path);
inode_t *path_inode (pid_t pid, const char *path);
inode_t *path_inode_link (pid_t pid, const char *path,
                          inode_t * inode, mode_t mode);
//----------------------------------------------------------
int fd_dup (pid_t pid, int fdn_old, int fdn_min);
fd_t *fd_reference (pid_t pid, int *fdn);
//----------------------------------------------------------
//
// void  sock_put (sock_t *s);
//
#define  sock_put(s)  (s->active=0)

sock_t *sock_reference (int skn);
h_port_t sock_free_port (void);
//----------------------------------------------------------

#endif
