#include <kernel/fs.h>
#include <errno.h>
#include <kernel/lib_k.h>
#include <kernel/lib_s.h>
//----------------------------------------------------------
inode_t *
inode_alloc (dev_t device, mode_t mode, uid_t uid,
             gid_t gid)
{
  sb_t *sb;
  inode_t *inode;
  int m;        // Index inside the inode map.
  int map_element;
  int map_bit;
  int map_mask;
  ino_t ino;
  // 
  // Check for arguments.
  // 
  if (mode == 0)
    {
      errset (EINVAL);  // Invalid argument.
      return (NULL);
    }
  // 
  // Get the super block from the known device.
  // 
  sb = sb_reference (device);
  if (sb == NULL)
    {
      errset (ENODEV);  // No such device.
      return (NULL);
    }
  // 
  // Find a free inode.
  // 
  while (1)
    {
      // 
      // Scan the inode bit map, to find a free inode
      // for new allocation.
      // 
      for (m = 0; m < (SB_MAP_INODE_SIZE * 16); m++)
        {
          map_element = m / 16;
          map_bit = m % 16;
          map_mask = 1 << map_bit;
          if (!(sb->map_inode[map_element] & map_mask))
            {
              // 
              // Found a free element: change the map
              // to
              // allocate the inode.
              // 
              sb->map_inode[map_element] |= map_mask;
              sb->changed = 1;
              ino = m;  // Found a free inode:
              break;    // exit the scan loop.
            }
        }
      // 
      // Check if the scan was successful.
      // 
      if (ino == 0)
        {
          errset (ENOSPC);      // No space left on
          // device.
          return (NULL);
        }
      // 
      // The inode was allocated inside the map in
      // memory.
      // 
      inode = inode_get (device, ino);
      if (inode == NULL)
        {
          errset (ENFILE);      // Too many files open 
          // in system.
          return (NULL);
        }
      // 
      // Verify if the inode is really free: if it
      // isn't, must save
      // it to disk.
      // 
      if (inode->size > 0 || inode->links > 0)
        {
          // 
          // Strange: should not have a size! Check if 
          // there are even
          // links. Please note that 255 links (that
          // is -1) is to be
          // considered a free inode, marked in a
          // special way for some
          // unknown reason. Currently, `LINK_MAX' is
          // equal to 254,
          // for that reason.
          // 
          if (inode->links > 0 && inode->links < LINK_MAX)
            {
              // 
              // Tell something.
              // 
              k_printf ("kernel alert: device %04x: "
                        "found \"free\" inode %i "
                        "that still has size %i "
                        "and %i links!\n",
                        device, ino, inode->size,
                        inode->links);
              // 
              // The inode must be set again to free,
              // inside
              // the bit map.
              // 
              map_element = ino / 16;
              map_bit = ino % 16;
              map_mask = 1 << map_bit;
              sb->map_inode[map_element] &= ~map_mask;
              sb->changed = 1;
              // 
              // Try to fix: reset all to zero.
              // 
              inode->mode = 0;
              inode->uid = 0;
              inode->gid = 0;
              inode->time = 0;
              inode->links = 0;
              inode->size = 0;
              inode->direct[0] = 0;
              inode->direct[1] = 0;
              inode->direct[2] = 0;
              inode->direct[3] = 0;
              inode->direct[4] = 0;
              inode->direct[5] = 0;
              inode->direct[6] = 0;
              inode->indirect1 = 0;
              inode->indirect2 = 0;
              inode->changed = 1;
              // 
              // Save fixed inode to disk.
              // 
              inode_put (inode);
              continue;
            }
          else
            {
              // 
              // Truncate the inode, save and break.
              // 
              inode_truncate (inode);
              inode_save (inode);
              break;
            }
        }
      else
        {
          // 
          // Considering free the inode found.
          // 
          break;
        }
    }
  // 
  // Put data inside the inode.
  // 
  inode->mode = mode;
  inode->uid = uid;
  inode->gid = gid;
  inode->size = 0;
  inode->time = s_time ((pid_t) 0, NULL);
  inode->links = 0;
  inode->changed = 1;
  // 
  // Save the inode.
  // 
  inode_save (inode);
  // 
  // Return the inode pointer.
  // 
  return (inode);
}
