#include <kernel/fs.h>
#include <errno.h>
#include <kernel/lib_k.h>
//----------------------------------------------------------
ssize_t
inode_file_read (inode_t * inode, off_t offset,
                 void *buffer, size_t count, int *eof)
{
  unsigned char *destination = (unsigned char *) buffer;
  unsigned char zone_buffer[SB_MAX_ZONE_SIZE];
  blkcnt_t blkcnt_read;
  off_t off_fzone;      // File zone offset.
  off_t off_buffer;     // Destination buffer offset.
  ssize_t size_read;    // Byte transfer counter.
  zno_t fzone;
  off_t off_end;
  // 
  // The inode pointer must be valid, and
  // the start byte must be positive.
  // 
  if (inode == NULL || offset < 0)
    {
      errset (EINVAL);  // Invalid argument.
      return ((ssize_t) - 1);
    }
  // 
  // Check if the start address is inside the file
  // size. This is not
  // an error, but zero bytes are read and `*eof' is
  // set. Otherwise,
  // `*eof' is reset.
  // 
  if (offset >= inode->size)
    {
      (eof != NULL) ? *eof = 1 : 0;
      return (0);
    }
  else
    {
      (eof != NULL) ? *eof = 0 : 0;
    }
  // 
  // Adjust, if necessary, the size of read, because
  // it cannot be
  // larger than the actual file size. The variable
  // `off_end' is
  // used to calculate the position *after* the
  // requested read.
  // Remember that the first file position is byte
  // zero; so,
  // the byte index inside the file goes from zero to
  // inode->size -1.
  // 
  off_end = offset;
  off_end += count;
  if (off_end > inode->size)
    {
      count = (inode->size - offset);
    }
  // 
  // Read the first file-zone inside the zone buffer.
  // 
  fzone = offset / inode->sb->blksize;
  off_fzone = offset % inode->sb->blksize;
  blkcnt_read =
    inode_fzones_read (inode, fzone, zone_buffer,
                       (blkcnt_t) 1);
  if (blkcnt_read <= 0)
    {
      // 
      // Sorry!
      // 

      k_printf
        ("inode_fzones_read (inode, fzone %i,... )\n",
         fzone);

      errset (EUNKNOWN);
      return (0);       // Zero bytes read!
    }
  // 
  // The first file-zone was read: copy it inside the
  // destination
  // buffer and continue reading the other zones
  // needed. Variables
  // `off_buffer' (destination buffer index) and
  // `size_read' (copy
  // byte counter) must be reset here. Variable
  // `off_fzone' is already
  // set with the initial offset inside `zone_buffer'.
  // 
  off_buffer = 0;
  size_read = 0;
  // 
  while (count)
    {
      // 
      // Copy the zone buffer into the destination.
      // Variables
      // `off_fzone', `off_buffer' and `size_read'
      // must not be
      // initialized inside the loop.
      // 
      for (;
           off_fzone < inode->sb->blksize && count > 0;
           off_fzone++, off_buffer++, size_read++,
           count--, offset++)
        {
          destination[off_buffer] = zone_buffer[off_fzone];
        }
      // 
      // If not all the bytes are copied, read the
      // next file-zone.
      // 
      if (count)
        {
          // 
          // Read another file-zone inside the zone
          // buffer.
          // Again, the function `inode_fzones_read()' 
          // might
          // return a null pointer, but the variable
          // `errno' tells if
          // it is really an error. For this reason,
          // the variable
          // `errno' must be reset before the read,
          // and checked after
          // it.
          // 
          fzone = offset / inode->sb->blksize;
          off_fzone = offset % inode->sb->blksize;
          blkcnt_read =
            inode_fzones_read (inode, fzone,
                               zone_buffer, (blkcnt_t) 1);
          if (blkcnt_read <= 0)
            {
              // 
              // Sorry: only `size_read' bytes read!
              // 
              errset (EUNKNOWN);
              return (size_read);
            }
        }
    }
  // 
  // The requested size was read completely.
  // 
  return (size_read);
}
