#include <kernel/fs.h>
#include <errno.h>
#include <kernel/lib_k.h>
//----------------------------------------------------------
ssize_t
inode_file_write (inode_t * inode, off_t offset,
                  const void *buffer, size_t count)
{
  unsigned char *buffer_source = (unsigned char *) buffer;
  unsigned char buffer_zone[SB_MAX_ZONE_SIZE];
  off_t off_fzone;      // File zone offset.
  off_t off_source;     // Source buffer offset.
  ssize_t size_copied;  // Byte transfer counter.
  ssize_t size_written; // Byte written counter.
  zno_t fzone;
  zno_t zone;
  blkcnt_t blkcnt_read;
  int status;
  // 
  // The inode pointer must be valid, and
  // the start byte must be positive.
  // 
  if (inode == NULL || offset < 0)
    {
      errset (EINVAL);  // Invalid argument.
      return ((ssize_t) - 1);
    }
  // 
  // Read a zone, modify it with the source buffer,
  // then write it back
  // and continue reading and writing other zones if
  // needed.
  // 
  for (size_written = 0, off_source = 0, size_copied =
       0; count > 0; size_written += size_copied)
    {
      // 
      // Read the next file-zone inside the zone
      // buffer: the function
      // `inode_zone()' is used to create
      // automatically the zone, if
      // it does not exist.
      // 
      fzone = offset / inode->sb->blksize;
      off_fzone = offset % inode->sb->blksize;
      zone = inode_zone (inode, fzone, 1);
      if (zone == 0)
        {
          // 
          // Return previously written bytes. The
          // variable `errno' is
          // already set by `inode_zone()'.
          // 
          return (size_written);
        }
      blkcnt_read =
        inode_fzones_read (inode, fzone, buffer_zone,
                           (blkcnt_t) 1);
      if (blkcnt_read <= 0)
        {
          // 
          // Even if the value is zero, there is a
          // problem reading the
          // zone to be overwritten (because
          // `inode_zone()' should
          // have already created such zone). The
          // variable `errno' is
          // already set by `inode_fzones_read()'.
          // 
          return ((ssize_t) - 1);
        }
      // 
      // The zone was successfully loaded inside the
      // buffer: overwrite
      // the zone buffer with the source buffer.
      // 
      for (size_copied = 0;
           off_fzone < inode->sb->blksize && count > 0;
           off_fzone++, off_source++, size_copied++,
           count--, offset++)
        {
          buffer_zone[off_fzone] =
            buffer_source[off_source];
        }
      // 
      // Save the zone.
      // 
      status = zone_write (inode->sb, zone, buffer_zone);
      if (status != 0)
        {
          // 
          // Cannot save the zone: return the size
          // already written.
          // The variable `errno' is already set by
          // `zone_write()'.
          // 
          return (size_written);
        }
      // 
      // Zone saved: update the file size if necessary 
      // (and the inode
      // too).
      // 
      if (inode->size <= offset)
        {
          inode->size = offset;
          inode->changed = 1;
          inode_save (inode);
        }
    }
  // 
  // All done successfully: return the value.
  // 
  return (size_written);
}
