#include <kernel/fs.h>
#include <errno.h>
#include <kernel/proc.h>
#include <libgen.h>
//----------------------------------------------------------
inode_t *
path_inode_link (pid_t pid, const char *path,
                 inode_t * inode, mode_t mode)
{
  proc_t *ps;
  char buffer[SB_MAX_ZONE_SIZE];
  off_t start;
  int d;        // Directory index.
  ssize_t size_read;
  ssize_t size_written;
  directory_t *dir = (directory_t *) buffer;
  char path_copy1[PATH_MAX];
  char path_copy2[PATH_MAX];
  char *path_directory;
  char *path_name;
  inode_t *inode_directory;
  inode_t *inode_new;
  dev_t device;
  int status;
  // 
  // Check arguments.
  // 
  if (path == NULL || strlen (path) == 0)
    {
      errset (EINVAL);  // Invalid argument:
      return (NULL);    // the path is mandatory.
    }
  // 
  if (inode == NULL && mode == 0)
    {
      errset (EINVAL);  // Invalid argument: if the
      // inode is to
      return (NULL);    // be created, the mode is
      // mandatory.
    }
  // 
  if (inode != NULL)
    {
      if (mode != 0)
        {
          errset (EINVAL);      // Invalid argument:
          // if the inode is
          return (NULL);        // already present,
          // the creation mode
        }       // must not be given.
      if (S_ISDIR (inode->mode))
        {
          errset (EPERM);       // Operation not
          // permitted.
          return (NULL);        // Refuse to link
          // directory.
        }
      if (inode->links >= LINK_MAX)
        {
          errset (EMLINK);      // Too many links.
          return (NULL);
        }
    }
  // 
  // Get process.
  // 
  ps = proc_reference (pid);
  // 
  // If the destination path already exists, the link
  // cannot be made.
  // It does not matter if the inode is known or not.
  // 
  inode_new = path_inode ((uid_t) 0, path);
  if (inode_new != NULL)
    {
      // 
      // A file already exists with the same name.
      // 
      inode_put (inode_new);
      errset (EEXIST);  // File exists.
      return (NULL);
    }
  // 
  // At this point, `inode_new' is `NULL'.
  // Copy the source path inside the directory path
  // and name arrays.
  // 
  strncpy (path_copy1, path, PATH_MAX);
  strncpy (path_copy2, path, PATH_MAX);
  // 
  // Reduce to directory name and find the last name.
  // 
  path_directory = dirname (path_copy1);
  path_name = basename (path_copy2);
  if (strlen (path_directory) == 0
      || strlen (path_name) == 0)
    {
      errset (EACCES);  // Permission denied: maybe
      // the
      // original path is the root directory
      // and cannot find a previous directory.
      return (NULL);
    }
  // 
  // Get the directory inode.
  // 
  inode_directory = path_inode (pid, path_directory);
  if (inode_directory == NULL)
    {
      errset (errno);
      return (NULL);
    }
  // 
  // Check if something is mounted on it.
  // 
  if (inode_directory->sb_attached != NULL)
    {
      // 
      // Must select the right directory.
      // 
      device = inode_directory->sb_attached->device;
      inode_put (inode_directory);
      inode_directory = inode_get (device, 1);
      if (inode_directory == NULL)
        {
          return (NULL);
        }
    }
  // 
  // If the inode to link is known, check if the
  // selected directory
  // has the same super block than the inode to link.
  // 
  if (inode != NULL && inode_directory->sb != inode->sb)
    {
      inode_put (inode_directory);
      errset (ENOENT);  // No such file or directory.
      return (NULL);
    }
  // 
  // Check if write is allowed for the file system.
  // 
  if (inode_directory->sb->options & MOUNT_RO)
    {
      inode_put (inode_directory);
      errset (EROFS);   // Read-only file system.
      return (NULL);
    }
  // 
  // Verify access permissions for the directory. The
  // number "3" means
  // that the user must have access permission and
  // write permission:
  // "-wx" == 2+1 == 3.
  // 
  status = inode_check (inode_directory, S_IFDIR, 3,
                        ps->euid, ps->egid);
  if (status != 0)
    {
      inode_put (inode_directory);
      return (NULL);
    }
  // 
  // If the inode to link was not specified, it must
  // be created.
  // From now on, the inode is referenced with the
  // variable
  // `inode_new'.
  // 
  inode_new = inode;
  // 
  if (inode_new == NULL)
    {
      inode_new =
        inode_alloc (inode_directory->sb->device, mode,
                     ps->euid, ps->egid);
      if (inode_new == NULL)
        {
          // 
          // The inode allocation failed, so, also the 
          // directory
          // must be released, before return.
          // 
          inode_put (inode_directory);
          return (NULL);
        }
    }
  // 
  // Read the directory content and try to add the new 
  // item.
  // 
  for (start = 0;
       start < inode_directory->size;
       start += inode_directory->sb->blksize)
    {
      size_read =
        inode_file_read (inode_directory, start,
                         buffer,
                         inode_directory->sb->blksize,
                         NULL);
      if (size_read < sizeof (directory_t))
        {
          break;
        }
      // 
      // Scan the directory portion just read, for an
      // unused item.
      // 
      dir = (directory_t *) buffer;
      for (d = 0; d < size_read;
           d += (sizeof (directory_t)), dir++)
        {
          if (dir->ino == 0)
            {
              // 
              // Found an empty directory item: link
              // the inode.
              // 
              dir->ino = inode_new->ino;
              strncpy (dir->name, path_name, NAME_MAX);
              inode_new->links++;
              inode_new->changed = 1;
              // 
              // Update the directory inside the file
              // system.
              // 
              size_written =
                inode_file_write (inode_directory,
                                  start, buffer, size_read);
              if (size_written != size_read)
                {
                  // 
                  // Write problem: release the
                  // directory and return.
                  // 
                  inode_put (inode_directory);
                  errset (EUNKNOWN);
                  return (NULL);
                }
              // 
              // Save the new inode, release the
              // directory and return
              // the linked inode.
              // 
              inode_save (inode_new);
              inode_put (inode_directory);
              return (inode_new);
            }
        }
    }
  // 
  // The directory don't have a free item and one must 
  // be appended.
  // 
  dir = (directory_t *) buffer;
  start = inode_directory->size;
  // 
  // Prepare the buffer with the link.
  // 
  dir->ino = inode_new->ino;
  strncpy (dir->name, path_name, NAME_MAX);
  inode_new->links++;
  inode_new->changed = 1;
  // 
  // Append the buffer to the directory.
  // 
  size_written =
    inode_file_write (inode_directory, start, buffer,
                      (sizeof (directory_t)));
  if (size_written != (sizeof (directory_t)))
    {
      // 
      // Problem updating the directory: release it
      // and return.
      // 
      inode_put (inode_directory);
      errset (EUNKNOWN);
      return (NULL);
    }
  // 
  // Close access to the directory inode and save the
  // other inode,
  // with updated link count.
  // 
  inode_put (inode_directory);
  inode_save (inode_new);
  // 
  // Return successfully.
  // 
  return (inode_new);
}
