#include <kernel/proc.h>
#include <kernel/lib_s.h>
#include <kernel/lib_k.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <arpa/inet.h>
//----------------------------------------------------------
int
s_bind (pid_t pid, int sfdn,
        const struct sockaddr *addr, socklen_t addrlen)
{
  fd_t *sfd;
  struct sockaddr_in *sin;
  proc_t *ps = proc_reference (pid);
  int i;
  clock_t clock_time;
  // 
  // Get file descriptor and verify that it is a
  // socket.
  // 
  sfd = fd_reference (pid, &sfdn);
  if (sfd == NULL || sfd->file == NULL)
    {
      errset (EBADF);   // Bad file descriptor.
      return (-1);
    }
  if (sfd->file->sock == NULL)
    {
      errset (ENOTSOCK);        // Not a socket.
      return (-1);
    }
  // 
  // Verify to have a valid address pointer.
  // 
  if (addr == NULL)
    {
      errset (EINVAL);
      return (-1);
    }
  // 
  // Check minimal address size.
  // 
  if (addrlen < sizeof (struct sockaddr))
    {
      errset (EINVAL);
      return (-1);
    }
  // 
  // 
  // 
  if (addr->sa_family == AF_INET)
    {
      sin = (struct sockaddr_in *) addr;
      // 
      // The source address might be zero, to tell
      // that any local
      // address is valid.
      // 
      // If it is a TCP/UDP protocol, must have valid
      // ports.
      // 
      if (sfd->file->sock->protocol == IPPROTO_TCP
          || sfd->file->sock->protocol == IPPROTO_UDP)
        {
          // 
          // Local port.
          // 
          if (ntohs (sin->sin_port) == 0)
            {
              // 
              // Missing the local port.
              // 
              errset (EADDRNOTAVAIL);
              return (-1);
            }
          // 
          // If the local port is privileged, must
          // have EUID == 0.
          // 
          if (ntohs (sin->sin_port) < 1024)
            {
              if (ps->euid != 0)
                {
                  // 
                  // Missing privileges.
                  // 
                  errset (EACCES);
                  return (-1);
                }
            }
          // 
          // If the local port is not given, a default 
          // one is assigned.
          // 
          if (sfd->file->sock->lport == 0)
            {
              // 
              // Must find a free one.
              // 
              sfd->file->sock->lport = sock_free_port ();
              if (sfd->file->sock->lport == 0)
                {
                  // 
                  // No port is available.
                  // 
                  errset (EAGAIN);
                  return (-1);
                }
            }
        }
      else
        {
          // 
          // Not supported.
          // 
          errset (EOPNOTSUPP);
          return (-1);
        }
      // 
      // Update the socket.
      // 
      sfd->file->sock->family = sin->sin_family;
      sfd->file->sock->laddr = ntohl (sin->sin_addr.s_addr);
      sfd->file->sock->lport = ntohs (sin->sin_port);
//      sfd->file->sock->bind   = 1;
      // 
      // Reset read packets clock time.
      // 
      clock_time = s_clock (pid);
      for (i = 0; i < IP_MAX_PACKETS; i++)
        {
          sfd->file->sock->read.clock[i] = clock_time;
        }
    }
  else
    {
      // 
      // Unsupported family.
      // 
      errset (EAFNOSUPPORT);
      return (-1);
    }
  // 
  // Ok.
  // 
  return (0);
}
