#include <kernel/proc.h>
#include <kernel/lib_s.h>
#include <kernel/lib_k.h>
#include <kernel/net/route.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <arpa/inet.h>
//----------------------------------------------------------
int
s_connect (pid_t pid, int sfdn,
           const struct sockaddr *addr, socklen_t addrlen)
{
  fd_t *sfd;
  struct sockaddr_in *sin;
  clock_t clock_time;
  int i;
  int status;
  // 
  // Get file descriptor and verify that it is a
  // socket.
  // 
  sfd = fd_reference (pid, &sfdn);
  if (sfd == NULL || sfd->file == NULL)
    {
      errset (EBADF);   // Bad file descriptor.
      return (-1);
    }
  if (sfd->file->sock == NULL)
    {
      errset (ENOTSOCK);        // Not a socket.
      return (-1);
    }
  // 
  // Verify to have a valid address pointer.
  // 
  if (addr == NULL)
    {
      errset (EINVAL);
      return (-1);
    }
  // 
  // Check minimal address size.
  // 
  if (addrlen < sizeof (struct sockaddr))
    {
      errset (EINVAL);
      return (-1);
    }
  // 
  // 
  // 
  if (addr->sa_family == AF_INET)
    {
      sin = (struct sockaddr_in *) addr;
      // 
      // Check to have the destination IP address.
      // 
      if (sin->sin_addr.s_addr == 0)
        {
          // 
          // This is not valid.
          // 
          errset (EADDRNOTAVAIL);
          return (-1);
        }
      // 
      // If it is a TCP/UDP protocol, must have valid
      // ports.
      // 
      if (sfd->file->sock->protocol == IPPROTO_TCP
          || sfd->file->sock->protocol == IPPROTO_UDP)
        {
          // 
          // Remote port.
          // 
          if (sin->sin_port == 0)
            {
              // 
              // Missing the remote port.
              // 
              errset (EADDRNOTAVAIL);
              return (-1);
            }
          // 
          // Local port.
          // 
          if (sfd->file->sock->lport == 0)
            {
              // 
              // Must find a free one.
              // 
              sfd->file->sock->lport = sock_free_port ();
              if (sfd->file->sock->lport == 0)
                {
                  // 
                  // No port is available.
                  // 
                  errset (EAGAIN);
                  return (-1);
                }
            }
        }
      // 
      // Update the socket, but not a TCP connection
      // already
      // working (TCP not connected has a zeroed
      // `tcp.conn' filled).
      // 
      if (sfd->file->sock->tcp.conn == 0
          || sfd->file->sock->tcp.conn == TCP_CLOSE)
        {
          // 
          // Update the socket.
          // 
          sfd->file->sock->family = sin->sin_family;
          sfd->file->sock->raddr =
            ntohl (sin->sin_addr.s_addr);
          sfd->file->sock->rport = ntohs (sin->sin_port);
          // 
          // Reset read packets clock time.
          // 
          clock_time = s_clock (pid);
          for (i = 0; i < IP_MAX_PACKETS; i++)
            {
              sfd->file->sock->read.clock[i] = clock_time;
            }
        }
      else
        {
          // 
          // It *is* a TCP connection already working: 
          // verify that
          // the socket is set as expected
          // 
          if (sfd->file->sock->family !=
              sin->sin_family
              || sfd->file->sock->raddr !=
              ntohl (sin->sin_addr.s_addr)
              || sfd->file->sock->rport !=
              ntohs (sin->sin_port))
            {
              // 
              // The socket address is changed!
              // 
              errset (EISCONN);
              return (-1);
            }
        }
      // 
      // If it is a TCP, not already working, should
      // connect.
      // The function `tcp_connect()' will verify the
      // connection
      // status.
      // 
      if (sfd->file->sock->protocol == IPPROTO_TCP)
        {
          // 
          // Be shure to have a source address.
          // 
          if (sfd->file->sock->laddr == 0)
            {
              // 
              // Default source address: get the
              // source address from the
              // routing table, based on the
              // destination.
              // 
              sfd->file->sock->laddr
                =
                route_remote_to_local (sfd->file->
                                       sock->raddr);
              if (sfd->file->sock->laddr ==
                  ((h_addr_t) - 1))
                {
                  errset (errno);
                  return (-1);
                }
            }
          // 
          // Call tcp_connect ().
          // 
          status = tcp_connect (sfd->file->sock);
          if (status)
            {
              errset (errno);
            }
          return (status);
        }
    }
  else
    {
      // 
      // It is not AF_INET: unsupported address
      // family.
      // 
      errset (EAFNOSUPPORT);
      return (-1);
    }
  // 
  // Ok.
  // 
  return (0);
}
