#ifndef _KERNEL_PROC_H
#define _KERNEL_PROC_H  1
//----------------------------------------------------------
#include <kernel/ibm_i386.h>
#include <kernel/dev.h>
#include <kernel/driver/tty.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <kernel/fs.h>
#include <sys/os32.h>
#include <stddef.h>
#include <stdint.h>
#include <time.h>
//----------------------------------------------------------
#define CLOCK_FREQUENCY_DIVISOR \
    (3579545/3/CLOCKS_PER_SEC)  // [1]
//
// [1] Internal clock frequency is (3579545/3) Hz.
//     This value is divided by
//     CLOCK_FREQUENCY_DIVISOR, giving 100 Hz.
//     The divisor value if fixed, because the code
//     suppose that the clock frequency is 100 Hz!
//
//----------------------------------------------------------
#define PROC_EMPTY                      0
#define PROC_CREATED                    1
#define PROC_READY                      2
#define PROC_RUNNING                    3
#define PROC_SLEEPING                   4
#define PROC_ZOMBIE                     5
//----------------------------------------------------------
#define MAGIC_OS32_APPL   0x6F7333326170706CLL  // [2]
//
// [2] os32appl
//
//----------------------------------------------------------
#define PROCESS_MAX  ((GDT_ITEMS/2)-1)  // Process
                                        // slots.
#define MAX_SIGNALS  31 // Max signal number.
//
typedef struct
{
  pid_t ppid;   // Parent PID.
  pid_t pgrp;   // Process group ID.
  uid_t uid;    // Real user ID
  uid_t euid;   // Effective user ID.
  uid_t suid;   // Saved user ID.
  gid_t gid;    // Real group ID
  gid_t egid;   // Effective group ID.
  gid_t sgid;   // Saved group ID.
  dev_t device_tty;     // Controlling terminal.
  char path_cwd[PATH_MAX];
  // Working directory path.
  inode_t *inode_cwd;   // Working directory inode.
  int umask;    // File creation mask.
  unsigned long int sig_status; // Active signals.
  unsigned long int sig_ignore; // Signals to be
  // ignored.
  uintptr_t sig_handler[MAX_SIGNALS];   // Opt. sig.
  // handlers.
  uintptr_t sig_handler_wrapper;        // Special
  // wrapper.
  clock_t usage;        // Clock ticks CPU time usage.
  unsigned int status;
  int wakeup_events;    // Wake up for something.
  int wakeup_signal;    // Signal waited.
  unsigned int wakeup_timer;    // Seconds to wait
  // for.
  inode_t *wakeup_inode;        // Inode waited.
  sock_t *wakeup_sock;  // Socket waited.
  dev_t wakeup_dev;     // Device waited.
  addr_t address_text;
  size_t domain_text;
  addr_t address_data;
  size_t domain_data;
  size_t domain_stack;  // Included inside the data.
  size_t extra_data;    // Extra data for `brk()'.
  uint32_t sp;
  int ret;
  char name[PATH_MAX];
  fd_t fd[FOPEN_MAX];
} proc_t;
//
extern proc_t proc_table[PROCESS_MAX];
//
// ATTENTION: THERE IS NO WAY TO KEEP THE STACK DOMAIN
//            IN A DIFFERENT ADDRESS SPACE THAN THE
//            OTHER DATA. There is a simple explanation
//            for such limitation: A POINTER TO DATA
//            INSIDE THE STACK, CANNOT BE REACHED IF IT
//            IS NOT INSIDE THE SAME ADDRESS SPACE!
//
//            For the same reason, data or stack,
//            cannot be moved (shifted) down, when more
//            space is needed, because previous
//            pointers would not work! So, to develop
//            an extensible `malloc' area, such memory
//            is to be placed *after* the stack, moving
//            *all* the data segment if necessary.
//----------------------------------------------------------
extern pid_t proc_current;
extern uint32_t proc_stack_pointer;
extern uint16_t proc_stack_segment_selector;
extern unsigned int proc_loops_per_clock;
//----------------------------------------------------------
typedef struct
{
  uint32_t filler0;
  uint64_t magic;
  uint32_t data_offset;
  uint32_t etext;
  uint32_t edata;
  uint32_t ebss;
  uint32_t ssize;
} header_t;
//----------------------------------------------------------
typedef struct
{
  uint32_t eax;
  uint32_t ecx;
  uint32_t edx;
  uint32_t ebx;
  uint32_t ebp;
  uint32_t esi;
  uint32_t edi;
  uint32_t ds;
  uint32_t es;
  uint32_t fs;
  uint32_t gs;
  uint32_t eip;
  uint32_t cs;
  uint32_t eflags;
} stack_t;
//----------------------------------------------------------
void proc_init (void);
void proc_timer_init (clock_t freq);
void proc_delay (void);
void proc_scheduler (void);
void sysroutine (uint32_t syscallnr, uint32_t msg_off,
                 uint32_t msg_size);
//----------------------------------------------------------
void *ptr (pid_t pid, void *p);
proc_t *proc_reference (pid_t pid);
void proc_print (void);
//----------------------------------------------------------
int proc_sys_exec (pid_t pid, const char *path,
                   unsigned int argc, char *arg_data,
                   unsigned int envc, char *env_data);
//----------------------------------------------------------
void proc_dump_memory (pid_t pid, addr_t address,
                       size_t size, char *name);
void proc_available (pid_t pid);
void proc_sch_net (void);
void proc_sch_signals (void);
void proc_sch_terminals (void);
void proc_sch_timers (void);
void proc_sig_chld (pid_t parent, int sig);
void proc_sig_cont (pid_t pid, int sig);
void proc_sig_core (pid_t pid, int sig);
void proc_sig_handler (pid_t pid, int sig);
int proc_sig_ignore (pid_t pid, int sig);
void proc_sig_off (pid_t pid, int sig);
void proc_sig_on (pid_t pid, int sig);
int proc_sig_status (pid_t pid, int sig);
void proc_sig_stop (pid_t pid, int sig);
void proc_sig_term (pid_t pid, int sig);

void proc_wakeup_pipe_read (inode_t * inode);
void proc_wakeup_pipe_write (inode_t * inode);
void proc_wakeup_terminal (void);

#endif
