#include <kernel/ibm_i386.h>
#include <kernel/proc.h>
#include <errno.h>
#include <fcntl.h>
#include <kernel/lib_s.h>
#include <kernel/lib_k.h>
//----------------------------------------------------------
#define DEBUG 0
//----------------------------------------------------------
int
proc_sys_exec (pid_t pid, const char *path,
               unsigned int argc, char *arg_data,
               unsigned int envc, char *env_data)
{
  unsigned int i;
  unsigned int j;
  char *arg;
  char *env;
  char *envp[ARG_MAX / 16];
  char *argv[ARG_MAX / 16];
  size_t size;
  size_t arg_data_size;
  size_t env_data_size;
  unsigned int p_off;
  inode_t *inode;
  ssize_t size_read;
  header_t header;
  uint32_t new_sp;
  uint32_t envp_address;
  uint32_t argv_address;
  addr_t allocated_text;
  addr_t allocated_data;
  addr_t stack_location;        // real stack
  // location.
  size_t process_domain_text;
  size_t process_domain_data;
  size_t process_domain_stack;
  addr_t previous_address_text;
  addr_t previous_address_data;
  size_t previous_domain_text;
  size_t previous_domain_data;
  size_t previous_domain_stack;
  size_t previous_extra_data;
  uint32_t segment_text;        // Segment descriptors
  uint32_t segment_data;        // inside 32 bit int.
  char buffer[MEM_BLOCK_SIZE];
  uint32_t stack_element;
  off_t off_inode;
  addr_t memory_start;
  int status;
  pid_t extra;
  int proc_count;
  file_t *file;
  int fdn;
  dev_t device;
  int eof;
  int sig;
  // 
  // Check for limits.
  // 
  if (argc > (ARG_MAX / 16) || envc > (ARG_MAX / 16))
    {
      errset (ENOMEM);
      return (-1);
    }
  // 
  // Scan arguments to calculate the full size and the 
  // relative
  // pointers. The final size is rounded to 4, for the 
  // stack.
  // 
  arg = arg_data;
  for (i = 0, j = 0; i < argc; i++)
    {
      argv[i] = (char *) j;     // Relative pointer
      // inside the
      // `arg_data'.
      size = strlen (arg);
      arg += size + 1;
      j += size + 1;
    }
  arg_data_size = j;
  if (arg_data_size % 2)
    {
      arg_data_size++;
    }
  if (arg_data_size % 4)
    {
      arg_data_size += 2;
    }
  // 
  // Scan environment variables to calculate the full
  // size and the
  // relative pointers. The final size is rounded to
  // 4, for the stack.
  // 
  env = env_data;
  for (i = 0, j = 0; i < envc; i++)
    {
      envp[i] = (char *) j;     // Relative pointer
      // inside the
      // `env_data'.
      size = strlen (env);
      env += size + 1;
      j += size + 1;
    }
  env_data_size = j;
  if (env_data_size % 2)
    {
      env_data_size++;
    }
  if (env_data_size % 4)
    {
      env_data_size += 2;
    }
  // 
  // Read the inode related to the executable file
  // name.
  // Function path_inode() includes the inode get
  // procedure.
  // 
  inode = path_inode (pid, path);
  if (inode == NULL)
    {
      errset (ENOENT);  // No such file or directory.
      return (-1);
    }
  // 
  // Check for permissions.
  // 
  status =
    inode_check (inode, S_IFREG, 5,
                 proc_table[pid].euid,
                 proc_table[pid].egid);
  if (status != 0)
    {
      // 
      // File is not of a valid type or permission are 
      // not
      // sufficient: release the executable file inode
      // and return with an error.
      // 
      inode_put (inode);
      errset (EACCES);  // Permission denied.
      return (-1);
    }
  // 
  // Read the header from the executable file.
  // 
  size_read =
    inode_file_read (inode, (off_t) 0, &header,
                     (sizeof header), &eof);
  if (size_read != (sizeof header))
    {
      // 
      // The file is shorter than the executable
      // header, so, it isn't
      // an executable: release the file inode and
      // return with an
      // error.
      // 
      inode_put (inode);
      errset (ENOEXEC);
      return (-1);
    }
  // 
  // Size read is ok.
  // 
  if (header.magic != MAGIC_OS32_APPL)
    {
      // 
      // The header does not have the expected magic
      // numbers, so,
      // it isn't a valid executable: release the file 
      // inode and
      // return with an error.
      // 
      inode_put (inode);
      errset (ENOEXEC);
      return (-1);      // This is not a valid
      // executable!
    }
  // 
  // Calculate code size.
  // 
  if (header.data_offset == 0)
    {
      process_domain_text = header.ebss + header.ssize;
    }
  else
    {
      process_domain_text = header.data_offset;
    }
  // 
  if (process_domain_text % 4096)
    {
      process_domain_text =
        (((process_domain_text / 4096) + 1) * 4096);
    }
  // 
  // Calculate data size, including stack, that cannot 
  // stay alone!
  // 
  process_domain_stack = header.ssize;
  // 
  if (header.data_offset == 0)
    {
      process_domain_data = 0;
    }
  else
    {
      process_domain_data = (header.ebss + header.ssize);
    }
  // 
  if (process_domain_data % 4096)
    {
      process_domain_data =
        (((process_domain_data / 4096) + 1) * 4096);
    }
  // 
  // Place the new stack pointer to the bottom of the
  // data area:
  // the stack pointer is relative to the data area,
  // so the last
  // relative position is equal to the size.
  // 
  if (header.data_offset == 0)
    {
      new_sp = process_domain_text;
    }
  else
    {
      new_sp = process_domain_data;
    }
  // 
  // Allocate memory: code and data.
  // 
  allocated_text = mb_alloc_size (process_domain_text);
  // 
  if (allocated_text == 0)
    {
      // 
      // The program instructions (code segment)
      // cannot be loaded
      // into memory: release the executable file
      // inode and return
      // with an error.
      // 
      inode_put (inode);
      errset (ENOMEM);  // Not enough space.
      return (-1);
    }
  else if (DEBUG)
    {
      k_printf ("%s:%i:mb_alloc_size(%zi)", __FILE__,
                __LINE__,
                (unsigned int) process_domain_text);
    }
  // 
  // 
  // 
  if (header.data_offset == 0)
    {
      // 
      // Code and data segments are the same: no need
      // to allocate more memory for the data segment.
      // 
      allocated_data = 0;
      process_domain_data = 0;
    }
  else
    {
      // 
      // Code and data segments are different: the
      // data
      // segment memory is allocated.
      // 
      allocated_data = mb_alloc_size (process_domain_data);
      if (allocated_data == 0)
        {
          // 
          // The separated program data (data segment) 
          // cannot be
          // loaded into memory: free the already
          // allocated memory
          // for the program instructions, release the 
          // executable
          // file inode and return with an error.
          // 
          mb_free (allocated_text, process_domain_text);
          if (DEBUG)
            {
              k_printf ("%s:%i:mb_free(%i, %zi)",
                        __FILE__, __LINE__,
                        (unsigned int) allocated_text,
                        process_domain_data);
            }
          inode_put (inode);
          errset (ENOMEM);      // Not enough space.
          return (-1);
        }
      else if (DEBUG)
        {
          k_printf ("%s:%i:mb_alloc_size(%zi)",
                    __FILE__, __LINE__,
                    process_domain_data);
        }
    }
  // 
  // Load executable in memory.
  // 
  if (header.data_offset == 0)
    {
      // 
      // Code and data share the same segment.
      // 
      for (eof = 0, memory_start = allocated_text,
           off_inode = 0, size_read = 0;
           off_inode < inode->size && !eof;
           off_inode += size_read)
        {
          memory_start += size_read;
          // 
          // Read a block of memory.
          // 
          size_read =
            inode_file_read (inode, off_inode, buffer,
                             MEM_BLOCK_SIZE, &eof);
          if (size_read < 0)
            {
              // 
              // Free memory and inode.
              // 
              mb_free (allocated_text, process_domain_text);
              if (DEBUG)
                {
                  k_printf ("%s:%i:mb_free(%i, %zi)",
                            __FILE__, __LINE__,
                            (unsigned int)
                            allocated_text,
                            process_domain_text);
                }
              inode_put (inode);
              errset (EIO);
              return (-1);
            }
          // 
          // Copy inside the right position to be
          // executed.
          // 
          dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE,
                  memory_start, buffer,
                  (size_t) size_read, NULL);
        }
    }
  else
    {
      // 
      // Code and data with different segments.
      // 
      for (eof = 0, memory_start = allocated_text,
           off_inode = 0, size_read = 0;
           off_inode < process_domain_text && !eof;
           off_inode += size_read)
        {
          memory_start += size_read;
          // 
          // Read a block of memory
          // 
          size_read =
            inode_file_read (inode, off_inode, buffer,
                             MEM_BLOCK_SIZE, &eof);
          if (size_read < 0)
            {
              // 
              // Free memory and inode.
              // 
              mb_free (allocated_text, process_domain_text);
              if (DEBUG)
                {
                  k_printf ("%s:%i:mb_free(%i, %zi)",
                            __FILE__, __LINE__,
                            (unsigned int)
                            allocated_text,
                            process_domain_text);
                }
              mb_free (allocated_data, process_domain_data);
              if (DEBUG)
                {
                  k_printf ("%s:%i:mb_free(%i, %zi)",
                            __FILE__, __LINE__,
                            (unsigned int)
                            allocated_data,
                            process_domain_data);
                }
              inode_put (inode);
              errset (EIO);
              return (-1);
            }
          // 
          // Copy inside the right position to be
          // executed.
          // 
          dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE,
                  memory_start, buffer,
                  (size_t) size_read, NULL);
        }
      for (eof = 0, memory_start = allocated_data,
           off_inode = header.data_offset, size_read =
           0; off_inode < inode->size && !eof;
           off_inode += size_read)
        {
          memory_start += size_read;
          // 
          // Read a block of memory
          // 
          size_read =
            inode_file_read (inode, off_inode, buffer,
                             MEM_BLOCK_SIZE, &eof);
          if (size_read < 0)
            {
              // 
              // Free memory and inode.
              // 
              mb_free (allocated_text, process_domain_text);
              if (DEBUG)
                {
                  k_printf ("%s:%i:mb_free(%i, %zi)",
                            __FILE__, __LINE__,
                            (unsigned int)
                            allocated_text,
                            process_domain_text);
                }
              mb_free (allocated_data, process_domain_data);
              if (DEBUG)
                {
                  k_printf ("%s:%i:mb_free(%i, %zi)",
                            __FILE__, __LINE__,
                            (unsigned int)
                            allocated_data,
                            process_domain_data);
                }
              inode_put (inode);
              errset (EIO);
              return (-1);
            }
          dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE,
                  memory_start, buffer,
                  (size_t) size_read, NULL);
        }
    }
  // 
  // The executable file was successfully loaded in
  // memory:
  // release the executable file inode.
  // 
  inode_put (inode);
  // 
  // Update process TEXT segment inside the GDT table.
  // 
  gdt_segment (gdt_pid_to_segment_text (pid),
               (uint32_t) allocated_text,
               (uint32_t) (process_domain_text / 4096),
               1, 1, 0);
  // 
  // Update process DATA segment inside the GDT table.
  // 
  if (process_domain_data > 0)
    {
      gdt_segment (gdt_pid_to_segment_data (pid),
                   (uint32_t) allocated_data,
                   (uint32_t) (process_domain_data /
                               4096), 1, 0, 0);
    }
  else
    {
      gdt_segment (gdt_pid_to_segment_data (pid),
                   (uint32_t) allocated_text,
                   (uint32_t) (process_domain_text /
                               4096), 1, 0, 0);
    }
  // 
  // Calculate segment descriptors.
  // 
  segment_text = (gdt_pid_to_segment_text (pid) << 3) + 0;
  segment_data = (gdt_pid_to_segment_data (pid) << 3) + 0;
  // 
  // Where is the stack?
  // 
  if (process_domain_data > 0)
    {
      stack_location = allocated_data;
    }
  else
    {
      stack_location = allocated_text;
    }
  // 
  // Put environment data inside the stack.
  // 
  new_sp -= env_data_size;      // --------------
  // ENVIRONMENT
  dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE,
          (off_t) (stack_location + new_sp),
          env_data, env_data_size, NULL);
  // 
  // Put arguments data inside the stack.
  // 
  new_sp -= arg_data_size;      // ----------------
  // ARGUMENTS
  dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE,
          (off_t) (stack_location + new_sp),
          arg_data, arg_data_size, NULL);
  // 
  // Put envp[] inside the stack, updating all the
  // pointers.
  // 
  new_sp -= 4;  // --------------------------------- NULL
  stack_element = (uint32_t) NULL;
  dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE,
          (off_t) (stack_location + new_sp),
          &stack_element, (sizeof stack_element), NULL);
  // 
  // Calculate memory pointers from original relative
  // pointers, inside the environment array of
  // pointers.
  // 
  p_off = new_sp;
  p_off += 4;
  p_off += arg_data_size;
  for (i = 0; i < envc; i++)
    {
      envp[i] += p_off;
    }
  // 
  new_sp -= (envc * (sizeof (char *))); // ----- *envp[]
  dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE,
          (off_t) (stack_location + new_sp),
          envp, (envc * (sizeof (char *))), NULL);
  // 
  // Save the envp[] location, needed in the
  // following.
  // 
  envp_address = new_sp;
  // 
  // Put argv[] inside the stack, updating all the
  // pointers.
  // 
  new_sp -= 4;  // -------------------------------- NULL
  stack_element = (uint32_t) NULL;
  dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE,
          (off_t) (stack_location + new_sp),
          &stack_element, (sizeof stack_element), NULL);
  // 
  // Calculate memory pointers from original relative
  // pointers, inside the arguments array of pointers.
  // 
  p_off = new_sp;
  p_off += 4;
  p_off += (envc * (sizeof (char *)));
  p_off += 4;
  for (i = 0; i < argc; i++)
    {
      argv[i] += p_off;
    }
  // 
  new_sp -= (argc * (sizeof (char *))); // ----- *argv[]
  dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE,
          (off_t) (stack_location + new_sp),
          argv, (argc * (sizeof (char *))), NULL);
  // 
  // Save the argv[] location, needed in the
  // following.
  // 
  argv_address = new_sp;
  // 
  // Put the pointer to the array envp[].
  // 
  new_sp -= 4;  // ------------------------------- argc
  stack_element = envp_address;
  dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE,
          (off_t) (stack_location + new_sp),
          &stack_element, (sizeof stack_element), NULL);
  // 
  // Put the pointer to the array argv[].
  // 
  new_sp -= 4;  // ------------------------------- argc
  stack_element = argv_address;
  dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE,
          (off_t) (stack_location + new_sp),
          &stack_element, (sizeof stack_element), NULL);
  // 
  // Put argc inside the stack.
  // 
  new_sp -= 4;  // ------------------------------- argc
  dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE,
          (off_t) (stack_location + new_sp),
          &argc, (sizeof argc), NULL);
  // 
  // Set the rest of the stack.
  // 
  new_sp -= 4;  // ----------------------------- EFLAGS
  stack_element = 0x0200;
  dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE,
          (off_t) (stack_location + new_sp),
          &stack_element, (sizeof stack_element), NULL);
  new_sp -= 4;  // --------------------------------- CS
  dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE,
          (off_t) (stack_location + new_sp),
          &segment_text, (sizeof segment_text), NULL);
  new_sp -= 4;  // -------------------------------- EIP
  stack_element = 0;
  dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE,
          (off_t) (stack_location + new_sp),
          &stack_element, (sizeof stack_element), NULL);
  new_sp -= 4;  // --------------------------------- GS
  dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE,
          (off_t) (stack_location + new_sp),
          &segment_data, (sizeof segment_data), NULL);
  new_sp -= 4;  // --------------------------------- FS
  dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE,
          (off_t) (stack_location + new_sp),
          &segment_data, (sizeof segment_data), NULL);
  new_sp -= 4;  // --------------------------------- ES
  dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE,
          (off_t) (stack_location + new_sp),
          &segment_data, (sizeof segment_data), NULL);
  new_sp -= 4;  // --------------------------------- DS
  dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE,
          (off_t) (stack_location + new_sp),
          &segment_data, (sizeof segment_data), NULL);
  new_sp -= 4;  // -------------------------------- EDI
  stack_element = 0;
  dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE,
          (off_t) (stack_location + new_sp),
          &stack_element, (sizeof stack_element), NULL);
  new_sp -= 4;  // -------------------------------- ESI
  stack_element = 0;
  dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE,
          (off_t) (stack_location + new_sp),
          &stack_element, (sizeof stack_element), NULL);
  new_sp -= 4;  // -------------------------------- EBP
  stack_element = 0;
  dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE,
          (off_t) (stack_location + new_sp),
          &stack_element, (sizeof stack_element), NULL);
  new_sp -= 4;  // -------------------------------- EBX
  stack_element = 0;
  dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE,
          (off_t) (stack_location + new_sp),
          &stack_element, (sizeof stack_element), NULL);
  new_sp -= 4;  // -------------------------------- EDX
  stack_element = 0;
  dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE,
          (off_t) (stack_location + new_sp),
          &stack_element, (sizeof stack_element), NULL);
  new_sp -= 4;  // -------------------------------- ECX
  stack_element = 0;
  dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE,
          (off_t) (stack_location + new_sp),
          &stack_element, (sizeof stack_element), NULL);
  new_sp -= 4;  // -------------------------------- EAX
  stack_element = 0;
  dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE,
          (off_t) (stack_location + new_sp),
          &stack_element, (sizeof stack_element), NULL);
  // 
  // Close process file descriptors, if the
  // `FD_CLOEXEC' flag
  // is present.
  // 
  for (fdn = 0; fdn < OPEN_MAX; fdn++)
    {
      if (proc_table[pid].fd[0].file != NULL)
        {
          if (proc_table[pid].fd[0].fd_flags & FD_CLOEXEC)
            {
              s_close (pid, fdn);
            }
        }
    }
  // 
  // Select device for standard I/O, if a standard I/O 
  // stream must be
  // opened.
  // 
  if (proc_table[pid].device_tty != 0)
    {
      device = proc_table[pid].device_tty;
    }
  else
    {
      device = DEV_TTY;
    }
  // 
  // Prepare missing standard file descriptors. The
  // function
  // `file_stdio_dev_make()' arranges the value for
  // `errno' if
  // necessary. If a standard file descriptor cannot
  // be allocated,
  // the program is left without it.
  // 
  if (proc_table[pid].fd[0].file == NULL)
    {
      file =
        file_stdio_dev_make (device, S_IFCHR, O_RDONLY);
      if (file != NULL) // stdin
        {
          proc_table[pid].fd[0].fl_flags = O_RDONLY;
          proc_table[pid].fd[0].fd_flags = 0;
          proc_table[pid].fd[0].file = file;
          proc_table[pid].fd[0].file->offset = 0;
        }
    }
  if (proc_table[pid].fd[1].file == NULL)
    {
      file =
        file_stdio_dev_make (device, S_IFCHR, O_WRONLY);
      if (file != NULL) // stdout
        {
          proc_table[pid].fd[1].fl_flags = O_WRONLY;
          proc_table[pid].fd[1].fd_flags = 0;
          proc_table[pid].fd[1].file = file;
          proc_table[pid].fd[1].file->offset = 0;
        }
    }
  if (proc_table[pid].fd[2].file == NULL)
    {
      file =
        file_stdio_dev_make (device, S_IFCHR, O_WRONLY);
      if (file != NULL) // stderr
        {
          proc_table[pid].fd[2].fl_flags = O_WRONLY;
          proc_table[pid].fd[2].fd_flags = 0;
          proc_table[pid].fd[2].file = file;
          proc_table[pid].fd[2].file->offset = 0;
        }
    }
  // 
  // Prepare to switch
  // 
  previous_address_text = proc_table[pid].address_text;
  previous_domain_text = proc_table[pid].domain_text;
  previous_address_data = proc_table[pid].address_data;
  previous_domain_data = proc_table[pid].domain_data;
  previous_domain_stack = proc_table[pid].domain_stack;
  previous_extra_data = proc_table[pid].extra_data;
  // 
  proc_table[pid].address_text = allocated_text;
  proc_table[pid].domain_text = process_domain_text;
  proc_table[pid].address_data = allocated_data;
  proc_table[pid].domain_data = process_domain_data;
  proc_table[pid].domain_stack = process_domain_stack;
  proc_table[pid].extra_data = (size_t) 0;
  proc_table[pid].sp = new_sp;
  strncpy (proc_table[pid].name, path, PATH_MAX);
  // 
  // Ensure to have a terminated string.
  // 
  proc_table[pid].name[PATH_MAX - 1] = 0;
  // 
  // Reset `sig_handler[]'.
  // 
  for (sig = 0; sig < MAX_SIGNALS; sig++)
    {
      proc_table[pid].sig_handler[sig] = (uintptr_t) NULL;
    }
  // 
  // Free previous data memory (included stack).
  // 
  if (previous_domain_data > 0)
    {
      mb_free (previous_address_data,
               previous_domain_data + previous_extra_data);
      if (DEBUG)
        {
          k_printf ("%s:%i:mb_free(%i, %zi)",
                    __FILE__, __LINE__,
                    (unsigned int)
                    previous_address_data,
                    previous_domain_data +
                    previous_extra_data);
        }
    }
  // 
  // Free code memory if not shared.
  // 
  for (proc_count = 0, extra = 0; extra < PROCESS_MAX;
       extra++)
    {
      if (proc_table[extra].status == PROC_EMPTY ||
          proc_table[extra].status == PROC_ZOMBIE)
        {
          continue;
        }
      if (previous_address_text ==
          proc_table[extra].address_text)
        {
          proc_count++;
        }
    }
  if (proc_count == 0)
    {
      // 
      // The code segment can be released, because no
      // other
      // process is using it.
      // 
      if (previous_domain_data > 0)
        {
          mb_free (previous_address_text,
                   previous_domain_text +
                   previous_extra_data);
          if (DEBUG)
            {
              k_printf ("%s:%i:mb_free(%i, %zi)",
                        __FILE__, __LINE__,
                        (unsigned int)
                        previous_address_text,
                        previous_domain_text +
                        previous_extra_data);
            }
        }
      else
        {
          mb_free (previous_address_text,
                   previous_domain_text);
          if (DEBUG)
            {
              k_printf ("%s:%i:mb_free(%i, %zi)",
                        __FILE__, __LINE__,
                        (unsigned int)
                        previous_address_text,
                        previous_domain_text);
            }
        }
    }
  // 
  // Change the segment and the stack pointer, from
  // the interrupt.
  // [1] Anyway,the stack segment selector does not
  // change.
  // 
  proc_stack_segment_selector = segment_data;   // [1]
  proc_stack_pointer = proc_table[pid].sp;
  // 
  // 
  // 
  return (0);
}
