#include <kernel/proc.h>
#include <errno.h>
#include <kernel/lib_k.h>
#include <kernel/lib_s.h>
#include <stdint.h>
//----------------------------------------------------------
static void sysroutine_error_back (int *number,
                                   int *line,
                                   char *file_name);
//----------------------------------------------------------
void
sysroutine (uint32_t syscallnr, uint32_t msg_off,
            uint32_t msg_size)
{
  pid_t pid = proc_current;
  // 
  // Inbox.
  // 
  union
  {
    sysmsg_accept_t accept;
    sysmsg_bind_t bind;
    sysmsg_brk_t brk;
    sysmsg_chdir_t chdir;
    sysmsg_chmod_t chmod;
    sysmsg_chown_t chown;
    sysmsg_clock_t clock;
    sysmsg_close_t close;
    sysmsg_connect_t connect;
    sysmsg_dup_t dup;
    sysmsg_dup2_t dup2;
    sysmsg_exec_t exec;
    sysmsg_exit_t exit;
    sysmsg_fchmod_t fchmod;
    sysmsg_fchown_t fchown;
    sysmsg_fcntl_t fcntl;
    sysmsg_fork_t fork;
    sysmsg_fstat_t fstat;
    sysmsg_ipconfig_t ipconfig;
    sysmsg_jmp_t jmp;
    sysmsg_kill_t kill;
    sysmsg_link_t link;
    sysmsg_listen_t listen;
    sysmsg_lseek_t lseek;
    sysmsg_mkdir_t mkdir;
    sysmsg_mknod_t mknod;
    sysmsg_mount_t mount;
    sysmsg_open_t open;
    sysmsg_pipe_t pipe;
    sysmsg_read_t read;
    sysmsg_recvfrom_t recvfrom;
    sysmsg_route_t route;
    sysmsg_send_t send;
    sysmsg_sbrk_t sbrk;
    sysmsg_seteuid_t seteuid;
    sysmsg_setuid_t setuid;
    sysmsg_setegid_t setegid;
    sysmsg_setgid_t setgid;
    sysmsg_signal_t signal;
    sysmsg_sleep_t sleep;
    sysmsg_socket_t socket;
    sysmsg_stat_t stat;
    sysmsg_stime_t stime;
    sysmsg_tcattr_t tcattr;
    sysmsg_time_t time;
    sysmsg_uarea_t uarea;
    sysmsg_umask_t umask;
    sysmsg_umount_t umount;
    sysmsg_unlink_t unlink;
    sysmsg_wait_t wait;
    sysmsg_write_t write;
    sysmsg_zpchar_t zpchar;
    sysmsg_zpstring_t zpstring;
  } *msg;
  // 
  // Align the message address pointer to the source
  // message.
  // 
  msg = ptr (pid, (void *) msg_off);
  // 
  // Verify if the system call was emitted from kernel 
  // code.
  // The kernel can emit only some particular system
  // call:
  // SYS_NULL, to let other processes run;
  // SYS_FORK, to let fork itself;
  // SYS_EXEC, to replace a forked copy of itself.
  // 
  if (pid == 0)
    {
      // 
      // This is the kernel code!
      // 
      if (syscallnr != SYS_0
          && syscallnr != SYS_FORK
          && syscallnr != SYS_EXEC
          && syscallnr != SYS_ZPSTRING)
        {
          k_printf
            ("kernel panic: the system call %i ",
             syscallnr);
          k_printf
            ("was received while running "
             "in kernel space!\n");
        }
    }
  // 
  // Entering a system call, the kernel variable
  // `errno' must be
  // reset, otherwise, a previous kernel code error
  // might be returned
  // to the applications.
  // 
  errno = 0;
  errln = 0;
  errfn[0] = 0;
  // 
  // Do the request from the received system call.
  // 
  switch (syscallnr)
    {
    case SYS_0:
      break;
    case SYS_ACCEPT:
      msg->accept.ret =
        s_accept (pid, msg->accept.sfdn,
                  &msg->accept.addr, &msg->accept.addrlen);
      msg->accept.fl_flags =
        proc_table[pid].fd[msg->accept.sfdn].fl_flags;
      sysroutine_error_back (&msg->accept.errno,
                             &msg->accept.errln,
                             msg->accept.errfn);
      break;
    case SYS_BIND:
      msg->bind.ret = s_bind (pid, msg->bind.sfdn,
                              &msg->bind.addr,
                              msg->bind.addrlen);
      sysroutine_error_back (&msg->bind.errno,
                             &msg->bind.errln,
                             msg->bind.errfn);
      break;
    case SYS_BRK:
      msg->brk.ret = s_brk (pid, msg->brk.address);
      sysroutine_error_back (&msg->brk.errno,
                             &msg->brk.errln,
                             msg->brk.errfn);
      break;
    case SYS_CHDIR:
      msg->chdir.ret =
        s_chdir (pid, ptr (pid, (void *) msg->chdir.path));
      sysroutine_error_back (&msg->chdir.errno,
                             &msg->chdir.errln,
                             msg->chdir.errfn);
      break;
    case SYS_CHMOD:
      msg->chmod.ret = s_chmod (pid,
                                ptr (pid,
                                     (void *) msg->
                                     chmod.path),
                                msg->chmod.mode);
      sysroutine_error_back (&msg->chmod.errno,
                             &msg->chmod.errln,
                             msg->chmod.errfn);
      break;
    case SYS_CHOWN:
      msg->chown.ret = s_chown (pid,
                                ptr (pid,
                                     (void *) msg->
                                     chown.path),
                                msg->chown.uid,
                                msg->chown.gid);
      sysroutine_error_back (&msg->chown.errno,
                             &msg->chown.errln,
                             msg->chown.errfn);
      break;
    case SYS_CLOCK:
      msg->clock.ret = s_clock (pid);
      break;
    case SYS_CLOSE:
      msg->close.ret = s_close (pid, msg->close.fdn);
      sysroutine_error_back (&msg->close.errno,
                             &msg->close.errln,
                             msg->close.errfn);
      break;
    case SYS_CONNECT:
      msg->connect.ret =
        s_connect (pid, msg->connect.sfdn,
                   &msg->connect.addr,
                   msg->connect.addrlen);
      sysroutine_error_back (&msg->connect.errno,
                             &msg->connect.errln,
                             msg->connect.errfn);
      break;
    case SYS_DUP:
      msg->dup.ret = s_dup (pid, msg->dup.fdn_old);
      sysroutine_error_back (&msg->dup.errno,
                             &msg->dup.errln,
                             msg->dup.errfn);
      break;
    case SYS_DUP2:
      msg->dup2.ret = s_dup2 (pid, msg->dup2.fdn_old,
                              msg->dup2.fdn_new);
      sysroutine_error_back (&msg->dup2.errno,
                             &msg->dup2.errln,
                             msg->dup2.errfn);
      break;
    case SYS_EXEC:
      msg->exec.ret = proc_sys_exec (pid,
                                     ptr (pid,
                                          (void *)
                                          msg->exec.path),
                                     msg->exec.argc,
                                     msg->exec.arg_data,
                                     msg->exec.envc,
                                     msg->exec.env_data);
      msg->exec.uid = proc_table[pid].uid;
      msg->exec.euid = proc_table[pid].euid;
      sysroutine_error_back (&msg->exec.errno,
                             &msg->exec.errln,
                             msg->exec.errfn);
      break;
    case SYS_EXIT:
      if (pid == 0)
        {
          k_printf ("kernel alert: "
                    "the kernel cannot exit!\n");
        }
      else
        {
          s__exit (pid, msg->exit.status);
        }
      break;
    case SYS_FCHMOD:
      msg->fchmod.ret = s_fchmod (pid, msg->fchmod.fdn,
                                  msg->fchmod.mode);
      sysroutine_error_back (&msg->fchmod.errno,
                             &msg->fchmod.errln,
                             msg->fchmod.errfn);
      break;
    case SYS_FCHOWN:
      msg->fchown.ret = s_fchown (pid, msg->fchown.fdn,
                                  msg->fchown.uid,
                                  msg->fchown.gid);
      sysroutine_error_back (&msg->fchown.errno,
                             &msg->fchown.errln,
                             msg->fchown.errfn);
      break;
    case SYS_FCNTL:
      msg->fcntl.ret = s_fcntl (pid, msg->fcntl.fdn,
                                msg->fcntl.cmd,
                                msg->fcntl.arg);
      sysroutine_error_back (&msg->fcntl.errno,
                             &msg->fcntl.errln,
                             msg->fcntl.errfn);
      break;
    case SYS_FORK:
      msg->fork.ret = s_fork (pid);
      sysroutine_error_back (&msg->fork.errno,
                             &msg->fork.errln,
                             msg->fork.errfn);
      break;
    case SYS_FSTAT:
      msg->fstat.ret =
        s_fstat (pid, msg->fstat.fdn, &msg->fstat.stat);
      sysroutine_error_back (&msg->fstat.errno,
                             &msg->fstat.errln,
                             msg->fstat.errfn);
      break;
    case SYS_IPCONFIG:
      msg->ipconfig.ret =
        s_ipconfig (pid, msg->ipconfig.n,
                    msg->ipconfig.address, msg->ipconfig.m);
      sysroutine_error_back (&msg->ipconfig.errno,
                             &msg->ipconfig.errln,
                             msg->ipconfig.errfn);
      break;
    case SYS_KILL:
      msg->kill.ret =
        s_kill (pid, msg->kill.pid, msg->kill.signal);
      sysroutine_error_back (&msg->kill.errno,
                             &msg->kill.errln,
                             msg->kill.errfn);
      break;
    case SYS_LINK:
      msg->link.ret
        = s_link (pid,
                  ptr (pid,
                       (void *) msg->link.path_old),
                  ptr (pid, (void *) msg->link.path_new));
      sysroutine_error_back (&msg->link.errno,
                             &msg->link.errln,
                             msg->link.errfn);
      break;
    case SYS_LISTEN:
      msg->listen.ret =
        s_listen (pid, msg->listen.sfdn,
                  msg->listen.backlog);
      sysroutine_error_back (&msg->listen.errno,
                             &msg->listen.errln,
                             msg->listen.errfn);
      break;
    case SYS_LONGJMP:
      s_longjmp (pid, msg->jmp.env, msg->jmp.ret);
      break;
    case SYS_LSEEK:
      msg->lseek.ret = s_lseek (pid, msg->lseek.fdn,
                                msg->lseek.offset,
                                msg->lseek.whence);
      sysroutine_error_back (&msg->lseek.errno,
                             &msg->lseek.errln,
                             msg->lseek.errfn);
      break;
    case SYS_MKDIR:
      msg->mkdir.ret = s_mkdir (pid,
                                ptr (pid,
                                     (void *) msg->
                                     mkdir.path),
                                msg->mkdir.mode);
      sysroutine_error_back (&msg->mkdir.errno,
                             &msg->mkdir.errln,
                             msg->mkdir.errfn);
      break;
    case SYS_MKNOD:
      msg->mknod.ret = s_mknod (pid,
                                ptr (pid,
                                     (void *) msg->
                                     mknod.path),
                                msg->mknod.mode,
                                msg->mknod.device);
      sysroutine_error_back (&msg->mknod.errno,
                             &msg->mknod.errln,
                             msg->mknod.errfn);
      break;
    case SYS_MOUNT:
      msg->mount.ret = s_mount (pid,
                                ptr (pid,
                                     (void *) msg->
                                     mount.path_dev),
                                ptr (pid,
                                     (void *) msg->
                                     mount.path_mnt),
                                msg->mount.options);
      sysroutine_error_back (&msg->mount.errno,
                             &msg->mount.errln,
                             msg->mount.errfn);
      break;
    case SYS_OPEN:
      msg->open.ret = s_open (pid,
                              ptr (pid,
                                   (void *) msg->open.path),
                              msg->open.flags,
                              msg->open.mode);
      sysroutine_error_back (&msg->open.errno,
                             &msg->open.errln,
                             msg->open.errfn);
      break;
    case SYS_PIPE:
      msg->pipe.ret = s_pipe (pid, msg->pipe.pipefd);
      sysroutine_error_back (&msg->pipe.errno,
                             &msg->pipe.errln,
                             msg->pipe.errfn);
      break;
    case SYS_PGRP:
      proc_table[pid].pgrp = pid;
      break;
    case SYS_READ:
      msg->read.ret = s_read (pid, msg->read.fdn,
                              ptr (pid,
                                   msg->read.buffer),
                              msg->read.count);
      msg->read.fl_flags =
        proc_table[pid].fd[msg->read.fdn].fl_flags;
      sysroutine_error_back (&msg->read.errno,
                             &msg->read.errln,
                             msg->read.errfn);
      break;
    case SYS_RECVFROM:
      msg->recvfrom.ret =
        s_recvfrom
        (pid, msg->recvfrom.sfdn,
         ptr (pid, msg->recvfrom.buffer),
         msg->recvfrom.count,
         msg->recvfrom.flags, ptr (pid,
                                   msg->recvfrom.addrfrom),
         ptr (pid, msg->recvfrom.addrsize));
      msg->recvfrom.fl_flags =
        proc_table[pid].fd[msg->recvfrom.sfdn].fl_flags;
      sysroutine_error_back (&msg->recvfrom.errno,
                             &msg->recvfrom.errln,
                             msg->recvfrom.errfn);
      break;
    case SYS_ROUTEADD:
      msg->route.ret =
        s_routeadd (pid, msg->route.destination,
                    msg->route.m, msg->route.router,
                    msg->route.device);
      sysroutine_error_back (&msg->route.errno,
                             &msg->route.errln,
                             msg->route.errfn);
      break;
    case SYS_ROUTEDEL:
      msg->route.ret =
        s_routedel (pid, msg->route.destination,
                    msg->route.m);
      sysroutine_error_back (&msg->route.errno,
                             &msg->route.errln,
                             msg->route.errfn);
      break;
    case SYS_SBRK:
      msg->sbrk.ret = s_sbrk (pid, msg->sbrk.increment);
      sysroutine_error_back (&msg->sbrk.errno,
                             &msg->sbrk.errln,
                             msg->sbrk.errfn);
      break;
    case SYS_SEND:
      msg->send.ret = s_send (pid, msg->send.sfdn,
                              ptr (pid,
                                   (void *) msg->send.
                                   buffer), msg->send.count,
                              msg->send.flags);
      sysroutine_error_back (&msg->send.errno,
                             &msg->send.errln,
                             msg->send.errfn);
      break;
    case SYS_SETEUID:
      msg->seteuid.ret = s_seteuid (pid, msg->seteuid.euid);
      msg->seteuid.euid = proc_table[pid].euid;
      sysroutine_error_back (&msg->seteuid.errno,
                             &msg->seteuid.errln,
                             msg->seteuid.errfn);
      break;
    case SYS_SETUID:
      msg->setuid.ret = s_setuid (pid, msg->setuid.euid);
      msg->setuid.uid = proc_table[pid].uid;
      msg->setuid.euid = proc_table[pid].euid;
      msg->setuid.suid = proc_table[pid].suid;
      sysroutine_error_back (&msg->setuid.errno,
                             &msg->setuid.errln,
                             msg->setuid.errfn);
      break;
    case SYS_SETEGID:
      msg->setegid.ret = s_setegid (pid, msg->setegid.egid);
      msg->setegid.egid = proc_table[pid].egid;
      sysroutine_error_back (&msg->setegid.errno,
                             &msg->setegid.errln,
                             msg->setegid.errfn);
      break;
    case SYS_SETGID:
      msg->setgid.ret = s_setgid (pid, msg->setgid.egid);
      msg->setgid.gid = proc_table[pid].gid;
      msg->setgid.egid = proc_table[pid].egid;
      msg->setgid.sgid = proc_table[pid].sgid;
      sysroutine_error_back (&msg->setgid.errno,
                             &msg->setgid.errln,
                             msg->setgid.errfn);
      break;
    case SYS_SETJMP:
      msg->jmp.ret = s_setjmp (pid, msg->jmp.env);
      break;
    case SYS_SIGNAL:
      msg->signal.ret =
        s_signal (pid, msg->signal.signal,
                  msg->signal.handler, msg->signal.wrapper);
      sysroutine_error_back (&msg->signal.errno,
                             &msg->signal.errln,
                             msg->signal.errfn);
      break;
    case SYS_SLEEP:
      proc_table[pid].status = PROC_SLEEPING;
      proc_table[pid].ret = 0;
      proc_table[pid].wakeup_events = msg->sleep.events;
      proc_table[pid].wakeup_timer = msg->sleep.seconds;
      break;
    case SYS_STAT:
      msg->stat.ret = s_stat (pid,
                              ptr (pid,
                                   (void *) msg->stat.path),
                              &msg->stat.stat);
      sysroutine_error_back (&msg->stat.errno,
                             &msg->stat.errln,
                             msg->stat.errfn);
      break;
    case SYS_STIME:
      msg->stime.ret = s_stime (pid, &msg->stime.timer);
      break;
    case SYS_TCGETATTR:
      msg->tcattr.ret =
        s_tcgetattr (pid, msg->tcattr.fdn,
                     ptr (pid, msg->tcattr.attr));
      sysroutine_error_back (&msg->tcattr.errno,
                             &msg->tcattr.errln,
                             msg->tcattr.errfn);
      break;
    case SYS_TCSETATTR:
      msg->tcattr.ret =
        s_tcsetattr (pid, msg->tcattr.fdn,
                     msg->tcattr.action, ptr (pid,
                                              msg->tcattr.
                                              attr));
      sysroutine_error_back (&msg->tcattr.errno,
                             &msg->tcattr.errln,
                             msg->tcattr.errfn);
      break;
    case SYS_TIME:
      msg->time.ret = s_time (pid, NULL);
      break;
    case SYS_UAREA:
      msg->uarea.uid = proc_table[pid].uid;
      msg->uarea.suid = proc_table[pid].suid;
      msg->uarea.euid = proc_table[pid].euid;
      msg->uarea.gid = proc_table[pid].gid;
      msg->uarea.sgid = proc_table[pid].sgid;
      msg->uarea.egid = proc_table[pid].egid;
      msg->uarea.pid = pid;
      msg->uarea.ppid = proc_table[pid].ppid;
      msg->uarea.pgrp = proc_table[pid].pgrp;
      msg->uarea.umask = proc_table[pid].umask;
      strncpy (ptr (pid, msg->uarea.path_cwd),
               proc_table[pid].path_cwd,
               msg->uarea.path_cwd_size);
      break;
    case SYS_UMASK:
      msg->umask.ret = proc_table[pid].umask;
      proc_table[pid].umask = (msg->umask.umask & 00777);
      break;
    case SYS_UMOUNT:
      msg->umount.ret = s_umount (pid,
                                  ptr (pid,
                                       (void *)
                                       msg->umount.
                                       path_mnt));
      sysroutine_error_back (&msg->umount.errno,
                             &msg->umount.errln,
                             msg->umount.errfn);
      break;
    case SYS_UNLINK:
      msg->unlink.ret = s_unlink (pid,
                                  ptr (pid,
                                       (void *) msg->
                                       unlink.path));
      sysroutine_error_back (&msg->unlink.errno,
                             &msg->unlink.errln,
                             msg->unlink.errfn);
      break;
    case SYS_WAIT:
      msg->wait.ret = s_wait (pid, &msg->wait.status);
      sysroutine_error_back (&msg->wait.errno,
                             &msg->wait.errln,
                             msg->wait.errfn);
      break;
    case SYS_WRITE:
      msg->write.ret = s_write (pid, msg->write.fdn,
                                ptr (pid,
                                     (void *) msg->
                                     write.buffer),
                                msg->write.count);
      sysroutine_error_back (&msg->write.errno,
                             &msg->write.errln,
                             msg->write.errfn);
      break;
    case SYS_SOCKET:
      msg->socket.ret =
        s_socket (pid, msg->socket.family,
                  msg->socket.type, msg->socket.protocol);
      sysroutine_error_back (&msg->socket.errno,
                             &msg->socket.errln,
                             msg->socket.errfn);
      break;
    case SYS_ZPCHAR:
      dev_io (pid, DEV_TTY, DEV_WRITE, 0L,
              &msg->zpchar.c, 1, NULL);
      break;
    case SYS_ZPSTRING:
      dev_io (pid, DEV_TTY, DEV_WRITE, (off_t) 0,
              msg->zpstring.string,
              strlen (msg->zpstring.string), NULL);
      break;
    default:
      k_printf
        ("kernel alert: unknown system call %i!\n",
         syscallnr);
      break;
    }
  // 
  // Continue with the scheduler.
  // 
  proc_scheduler ();
}

//----------------------------------------------------------
static void
sysroutine_error_back (int *number, int *line,
                       char *file_name)
{
  *number = errno;
  *line = errln;
  strncpy (file_name, errfn, PATH_MAX);
  file_name[PATH_MAX - 1] = 0;
}
